
'''
Ex23 from PETSc example files implemented for PETSc4py.
https://petsc.org/release/src/ksp/ksp/tutorials/ex23.c.html
By: Miguel Arriaga

Solves a tridiagonal linear system.

Vec            x, b, u;           approx solution, RHS, exact solution
Mat            A;                 linear system matrix
KSP            ksp;               linear solver context
PC             pc;                preconditioner context
PetscReal      norm;              norm of solution error

'''
import sys
import petsc4py
petsc4py.init(sys.argv)
from petsc4py import PETSc

import numpy as np

comm = PETSc.COMM_WORLD
size = comm.getSize()
rank = comm.getRank()
n = 12 # Size of problem
tol = 1E-11 # Tolerance of Result. tol=1000.*PETSC_MACHINE_EPSILON;

'''
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    Compute the matrix and right-hand-side vector that define
    the linear system, Ax = b.
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

Create vectors.  Note that we form 1 vector from scratch and
then duplicate as needed. For this simple case let PETSc decide how
many elements of the vector are stored on each processor. The second
argument to VecSetSizes() below causes PETSc to decide.
'''

x = PETSc.Vec().create(comm=comm)
x.setSizes(n)
x.setFromOptions()

b = x.duplicate()
u = x.duplicate()

'''
Identify the starting and ending mesh points on each
processor for the interior part of the mesh. We let PETSc decide
above.
'''

rstart,rend = x.getOwnershipRange()
nlocal = x.getLocalSize()

'''
Create matrix.  When using MatCreate(), the matrix format can
be specified at runtime.

Performance tuning note:  For problems of substantial size,
preallocation of matrix memory is crucial for attaining good
performance. See the matrix chapter of the users manual for details.

We pass in nlocal as the "local" size of the matrix to force it
to have the same parallel layout as the vector created above.
'''

A = PETSc.Mat().create(comm=comm)
A.setSizes(n,nlocal)
A.setFromOptions()
A.setUp()

'''
Assemble matrix.

The linear system is distributed across the processors by
chunks of contiguous rows, which correspond to contiguous
sections of the mesh on which the problem is discretized.
For matrix assembly, each processor contributes entries for
the part that it owns locally.
'''

col = np.zeros(3,dtype=PETSc.IntType)
value = np.zeros(3,dtype=PETSc.ScalarType)

if not rstart:
    rstart = 1
    i = 0; col[0] = 0; col[1] = 1; value[0] = 2.0; value[1] = -1.0
    A.setValues(i,col[0:2],value[0:2])

if rend == n:
    rend = n-1
    i = n-1; col[0] = n-2; col[1] = n-1; value[0] = -1.0; value[1] = 2.0
    A.setValues(i,col[0:2],value[0:2])


''' Set entries corresponding to the mesh interior '''
value[0] = -1.0; value[1] = 2.0; value[2] = -1.0
for i in range(rstart,rend):
    col[0] = i-1; col[1] = i; col[2] = i+1
    A.setValues(i,col,value)


''' Assemble the matrix '''
A.assemblyBegin(A.AssemblyType.FINAL)
A.assemblyEnd(A.AssemblyType.FINAL)

''' Set exact solution; then compute right-hand-side vector. '''
u.set(1.0)
b = A(u)

'''
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
        Create the linear solver and set various options
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - '''
'''
Create linear solver context
'''
ksp = PETSc.KSP().create()

'''
Set operators. Here the matrix that defines the linear system
also serves as the matrix from which the preconditioner is constructed.
'''
ksp.setOperators(A,A)

'''
Set linear solver defaults for this problem (optional).
    - By extracting the KSP and PC contexts from the KSP context,
      we can then directly call any KSP and PC routines to set
      various options.
    - The following four statements are optional; all of these
      parameters could alternatively be specified at runtime via
      KSPSetFromOptions();
'''
pc = ksp.getPC()
pc.setType('jacobi')
ksp.setTolerances(rtol=1.e-7)

'''
Set runtime options, e.g.,
-ksp_type <type> -pc_type <type> -ksp_monitor -ksp_rtol <rtol>
These options will override those specified above as long as
KSPSetFromOptions() is called _after_ any other customization
routines.
'''
ksp.setFromOptions()

'''
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
                Solve the linear system
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - '''

''' Solve linear system '''
ksp.solve(b,x)

'''
View solver info; we could instead use the option -ksp_view to
print this info to the screen at the conclusion of KSPSolve().
'''
# ksp.view()

'''
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
                Check solution and clean up
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - '''

''' Check the error '''
x = x - u # x.axpy(-1.0,u)
norm = x.norm(PETSc.NormType.NORM_2)
its = ksp.getIterationNumber()
if norm > tol:
    PETSc.Sys.Print("Norm of error {}, Iterations {}\n".format(norm,its),comm=comm)
else:
    if size==1:
        PETSc.Sys.Print("- Serial OK",comm=comm)
    else:
        PETSc.Sys.Print("- Parallel OK",comm=comm)



cdef extern from * nogil:
    PetscErrorCode PetscMalloc(size_t, void*)
    PetscErrorCode PetscFree(void*)
    PetscErrorCode PetscMemcpy(void*, void*, size_t)
    PetscErrorCode PetscMemmove(void*, void*, size_t)
    PetscErrorCode PetscMemzero(void*, size_t)
    PetscErrorCode PetscMemcmp(void*, void*, size_t, PetscBool*)
    PetscErrorCode PetscStrallocpy(const char[], char*[])



import petsc4py
from petsc4py import PETSc
import unittest
import os
import filecmp
import numpy as np
import importlib

# --------------------------------------------------------------------

ERR_ARG_OUTOFRANGE = 63


class BaseTestPlex:
    COMM = PETSc.COMM_WORLD
    DIM = 1
    CELLS = [[0, 1], [1, 2]]
    COORDS = [[0.0], [0.5], [1.0]]
    COMP = 1
    DOFS = [1, 0]

    def setUp(self):
        self.plex = PETSc.DMPlex().createFromCellList(
            self.DIM, self.CELLS, self.COORDS, comm=self.COMM
        )

    def tearDown(self):
        self.plex.destroy()
        self.plex = None
        PETSc.garbage_cleanup()

    def testTopology(self):
        rank = self.COMM.rank
        dim = self.plex.getDimension()
        pStart, pEnd = self.plex.getChart()
        cStart, cEnd = self.plex.getHeightStratum(0)
        vStart, vEnd = self.plex.getDepthStratum(0)
        numDepths = self.plex.getLabelSize('depth')
        coords_raw = self.plex.getCoordinates().getArray()
        coords = np.reshape(coords_raw, (vEnd - vStart, dim))
        self.assertEqual(dim, self.DIM)
        self.assertEqual(numDepths, self.DIM + 1)
        if rank == 0 and self.CELLS is not None:
            self.assertEqual(cEnd - cStart, len(self.CELLS))
        if rank == 0 and self.COORDS is not None:
            self.assertEqual(vEnd - vStart, len(self.COORDS))
            self.assertTrue((coords == self.COORDS).all())

    def testClosure(self):
        pStart, pEnd = self.plex.getChart()
        for p in range(pStart, pEnd):
            closure = self.plex.getTransitiveClosure(p)[0]
            for c in closure:
                cone = self.plex.getCone(c)
                self.assertEqual(self.plex.getConeSize(c), len(cone))
                for i in cone:
                    self.assertIn(i, closure)
            star = self.plex.getTransitiveClosure(p, useCone=False)[0]
            for s in star:
                support = self.plex.getSupport(s)
                self.assertEqual(self.plex.getSupportSize(s), len(support))
                for i in support:
                    self.assertIn(i, star)

    def testAdjacency(self):
        PETSc.DMPlex.setAdjacencyUseAnchors(self.plex, False)
        flag = PETSc.DMPlex.getAdjacencyUseAnchors(self.plex)
        self.assertFalse(flag)
        PETSc.DMPlex.setAdjacencyUseAnchors(self.plex, True)
        flag = PETSc.DMPlex.getAdjacencyUseAnchors(self.plex)
        self.assertTrue(flag)
        PETSc.DMPlex.setBasicAdjacency(self.plex, False, False)
        flagA, flagB = PETSc.DMPlex.getBasicAdjacency(self.plex)
        self.assertFalse(flagA)
        self.assertFalse(flagB)
        PETSc.DMPlex.setBasicAdjacency(self.plex, True, True)
        flagA, flagB = PETSc.DMPlex.getBasicAdjacency(self.plex)
        self.assertTrue(flagA)
        self.assertTrue(flagB)
        pStart, pEnd = self.plex.getChart()
        for p in range(pStart, pEnd):
            adjacency = self.plex.getAdjacency(p)
            self.assertTrue(p in adjacency)
            self.assertTrue(len(adjacency) > 1)

    def testSectionDofs(self):
        self.plex.setNumFields(1)
        section = self.plex.createSection([self.COMP], [self.DOFS])
        size = section.getStorageSize()
        entity_dofs = [
            self.plex.getStratumSize('depth', d) * self.DOFS[d]
            for d in range(self.DIM + 1)
        ]
        self.assertEqual(sum(entity_dofs), size)

    def testSectionClosure(self):
        section = self.plex.createSection([self.COMP], [self.DOFS])
        self.plex.setSection(section)
        vec = self.plex.createLocalVec()
        pStart, pEnd = self.plex.getChart()
        for p in range(pStart, pEnd):
            for i in range(section.getDof(p)):
                off = section.getOffset(p)
                vec.setValue(off + i, p)

        for p in range(pStart, pEnd):
            point_closure = self.plex.getTransitiveClosure(p)[0]
            dof_closure = self.plex.vecGetClosure(section, vec, p)
            for p in dof_closure:
                self.assertIn(p, point_closure)

    def testBoundaryLabel(self):
        pStart, pEnd = self.plex.getChart()
        if pEnd - pStart == 0:
            return

        self.assertFalse(self.plex.hasLabel('boundary'))
        self.plex.markBoundaryFaces('boundary')
        self.assertTrue(self.plex.hasLabel('boundary'))

        faces = self.plex.getStratumIS('boundary', 1)
        for f in faces.getIndices():
            points, orient = self.plex.getTransitiveClosure(f, useCone=True)
            for p in points:
                self.plex.setLabelValue('boundary', p, 1)

        for p in range(pStart, pEnd):
            if self.plex.getLabelValue('boundary', p) != 1:
                self.plex.setLabelValue('boundary', p, 2)

        numBoundary = self.plex.getStratumSize('boundary', 1)
        numInterior = self.plex.getStratumSize('boundary', 2)
        self.assertNotEqual(numBoundary, pEnd - pStart)
        self.assertNotEqual(numInterior, pEnd - pStart)
        self.assertEqual(numBoundary + numInterior, pEnd - pStart)

    def testMetric(self):
        if self.DIM == 1:
            return
        self.plex.distribute()
        if self.CELLS is None and not self.plex.isSimplex():
            return
        self.plex.orient()

        h_min = 1.0e-30
        h_max = 1.0e30
        a_max = 1.0e10
        target = 8.0
        p = 1.0
        beta = 1.3
        hausd = 0.01
        self.plex.metricSetUniform(False)
        self.plex.metricSetIsotropic(False)
        self.plex.metricSetRestrictAnisotropyFirst(False)
        self.plex.metricSetNoInsertion(False)
        self.plex.metricSetNoSwapping(False)
        self.plex.metricSetNoMovement(False)
        self.plex.metricSetNoSurf(False)
        self.plex.metricSetVerbosity(-1)
        self.plex.metricSetNumIterations(3)
        self.plex.metricSetMinimumMagnitude(h_min)
        self.plex.metricSetMaximumMagnitude(h_max)
        self.plex.metricSetMaximumAnisotropy(a_max)
        self.plex.metricSetTargetComplexity(target)
        self.plex.metricSetNormalizationOrder(p)
        self.plex.metricSetGradationFactor(beta)
        self.plex.metricSetHausdorffNumber(hausd)

        self.assertFalse(self.plex.metricIsUniform())
        self.assertFalse(self.plex.metricIsIsotropic())
        self.assertFalse(self.plex.metricRestrictAnisotropyFirst())
        self.assertFalse(self.plex.metricNoInsertion())
        self.assertFalse(self.plex.metricNoSwapping())
        self.assertFalse(self.plex.metricNoMovement())
        self.assertFalse(self.plex.metricNoSurf())
        self.assertTrue(self.plex.metricGetVerbosity() == -1)
        self.assertTrue(self.plex.metricGetNumIterations() == 3)
        self.assertTrue(np.isclose(self.plex.metricGetMinimumMagnitude(), h_min))
        self.assertTrue(np.isclose(self.plex.metricGetMaximumMagnitude(), h_max))
        self.assertTrue(np.isclose(self.plex.metricGetMaximumAnisotropy(), a_max))
        self.assertTrue(np.isclose(self.plex.metricGetTargetComplexity(), target))
        self.assertTrue(np.isclose(self.plex.metricGetNormalizationOrder(), p))
        self.assertTrue(np.isclose(self.plex.metricGetGradationFactor(), beta))
        self.assertTrue(np.isclose(self.plex.metricGetHausdorffNumber(), hausd))

        metric1 = self.plex.metricCreateUniform(0.5)
        metric2 = self.plex.metricCreateUniform(1.0)
        metric = self.plex.metricCreate()
        det = self.plex.metricDeterminantCreate()
        self.plex.metricAverage2(metric1, metric2, metric)
        metric1.array[:] *= 1.5
        self.assertTrue(np.allclose(metric.array, metric1.array))
        self.plex.metricIntersection2(metric1, metric2, metric)
        self.assertTrue(np.allclose(metric.array, metric2.array))
        self.plex.metricEnforceSPD(metric, metric1, det[0])
        self.assertTrue(np.allclose(metric.array, metric1.array))

        if self.DIM == 2 and PETSc.COMM_WORLD.getSize() > 6:
            # Error with 7 processes in 2D: normalization factor is -1
            return

        self.plex.metricNormalize(
            metric, metric1, det[0], restrictSizes=False, restrictAnisotropy=False
        )
        metric2.scale(pow(target, 2.0 / self.DIM))
        self.assertTrue(np.allclose(metric1.array, metric2.array))

    def testAdapt(self):
        if self.DIM == 1:
            return
        if self.DIM == 3 and PETSc.COMM_WORLD.getSize() > 4:
            # Error with 5 processes in 3D
            # ----------------------------
            # Warning: MMG5_mmgIntextmet: Unable to diagonalize at least 1 metric.
            # Error: MMG3D_defsiz_ani: unable to intersect metrics at point 8.
            # Metric undefined. Exit program.
            # MMG remeshing problem. Exit program.
            return
        self.plex.orient()
        plex = self.plex.refine()
        plex.distribute()
        if self.CELLS is None and not plex.isSimplex():
            return
        if sum(self.DOFS) > 1:
            return
        metric = plex.metricCreateUniform(9.0)
        try:
            newplex = plex.adaptMetric(metric, '')
            plex.destroy()
            newplex.destroy()
        except PETSc.Error as exc:
            plex.destroy()
            if exc.ierr != ERR_ARG_OUTOFRANGE:
                raise

    def testNatural(self):
        dim = self.plex.getDimension()
        ct = self.plex.getCellType(0)
        fe = PETSc.FE().createByCell(dim, 1, ct)
        self.plex.setField(0, fe)
        self.plex.createDS()
        self.plex.setUseNatural(True)
        self.plex.distribute()
        self.plex.view()
        gv = self.plex.createGlobalVec()
        nv = self.plex.createNaturalVec()
        self.plex.globalToNaturalBegin(gv, nv)
        self.plex.globalToNaturalEnd(gv, nv)
        self.plex.naturalToGlobalBegin(nv, gv)
        self.plex.naturalToGlobalEnd(nv, gv)

# --------------------------------------------------------------------


class BaseTestPlex_2D(BaseTestPlex):
    DIM = 2
    CELLS = [
        [0, 1, 3],
        [1, 3, 4],
        [1, 2, 4],
        [2, 4, 5],
        [3, 4, 6],
        [4, 6, 7],
        [4, 5, 7],
        [5, 7, 8],
    ]
    COORDS = [
        [0.0, 0.0],
        [0.5, 0.0],
        [1.0, 0.0],
        [0.0, 0.5],
        [0.5, 0.5],
        [1.0, 0.5],
        [0.0, 1.0],
        [0.5, 1.0],
        [1.0, 1.0],
    ]
    DOFS = [1, 0, 0]


class BaseTestPlex_3D(BaseTestPlex):
    DIM = 3
    CELLS = [
        [0, 2, 3, 7],
        [0, 2, 6, 7],
        [0, 4, 6, 7],
        [0, 1, 3, 7],
        [0, 1, 5, 7],
        [0, 4, 5, 7],
    ]
    COORDS = [
        [0.0, 0.0, 0.0],
        [1.0, 0.0, 0.0],
        [0.0, 1.0, 0.0],
        [1.0, 1.0, 0.0],
        [0.0, 0.0, 1.0],
        [1.0, 0.0, 1.0],
        [0.0, 1.0, 1.0],
        [1.0, 1.0, 1.0],
    ]
    DOFS = [1, 0, 0, 0]


# --------------------------------------------------------------------


class TestPlex_1D(BaseTestPlex, unittest.TestCase):
    pass


class TestPlex_2D(BaseTestPlex_2D, unittest.TestCase):
    def testTransform(self):
        plex = self.plex
        cstart, cend = plex.getHeightStratum(0)
        tr = PETSc.DMPlexTransform().create(comm=PETSc.COMM_WORLD)
        tr.setType(PETSc.DMPlexTransformType.REFINEALFELD)
        tr.setDM(plex)
        tr.setUp()
        newplex = tr.apply(plex)
        tr.destroy()
        newcstart, newcend = newplex.getHeightStratum(0)
        newplex.destroy()
        self.assertTrue((newcend - newcstart) == 3 * (cend - cstart))


class TestPlex_3D(BaseTestPlex_3D, unittest.TestCase):
    pass


class TestPlex_2D_P3(BaseTestPlex_2D, unittest.TestCase):
    DOFS = [1, 2, 1]


class TestPlex_3D_P3(BaseTestPlex_3D, unittest.TestCase):
    DOFS = [1, 2, 1, 0]


class TestPlex_3D_P4(BaseTestPlex_3D, unittest.TestCase):
    DOFS = [1, 3, 3, 1]


class TestPlex_2D_BoxTensor(BaseTestPlex_2D, unittest.TestCase):
    CELLS = None
    COORDS = None

    def setUp(self):
        self.plex = PETSc.DMPlex().createBoxMesh([3, 3], simplex=False)


class TestPlex_3D_BoxTensor(BaseTestPlex_3D, unittest.TestCase):
    CELLS = None
    COORDS = None

    def setUp(self):
        self.plex = PETSc.DMPlex().createBoxMesh([3, 3, 3], simplex=False)


# FIXME
try:
    raise PETSc.Error
    PETSc.DMPlex().createBoxMesh([2, 2], simplex=True, comm=PETSc.COMM_SELF).destroy()
except PETSc.Error:
    pass
else:

    class TestPlex_2D_Box(BaseTestPlex_2D, unittest.TestCase):
        CELLS = None
        COORDS = None

        def setUp(self):
            self.plex = PETSc.DMPlex().createBoxMesh([1, 1], simplex=True)

    class TestPlex_2D_Boundary(BaseTestPlex_2D, unittest.TestCase):
        CELLS = None
        COORDS = None

        def setUp(self):
            boundary = PETSc.DMPlex().create(self.COMM)
            boundary.createSquareBoundary([0.0, 0.0], [1.0, 1.0], [2, 2])
            boundary.setDimension(self.DIM - 1)
            self.plex = PETSc.DMPlex().generate(boundary)

    class TestPlex_3D_Box(BaseTestPlex_3D, unittest.TestCase):
        CELLS = None
        COORDS = None

        def setUp(self):
            self.plex = PETSc.DMPlex().createBoxMesh([1, 1, 1], simplex=True)

    class TestPlex_3D_Boundary(BaseTestPlex_3D, unittest.TestCase):
        CELLS = None
        COORDS = None

        def setUp(self):
            boundary = PETSc.DMPlex().create(self.COMM)
            boundary.createCubeBoundary([0.0, 0.0, 0.0], [1.0, 1.0, 1.0], [1, 1, 1])
            boundary.setDimension(self.DIM - 1)
            self.plex = PETSc.DMPlex().generate(boundary)

# --------------------------------------------------------------------

PETSC_DIR = petsc4py.get_config()['PETSC_DIR']


def check_dtype(method):
    def wrapper(self, *args, **kwargs):
        if PETSc.ScalarType is PETSc.ComplexType:
            return None
        return method(self, *args, **kwargs)

    return wrapper


def check_package(method):
    def wrapper(self, *args, **kwargs):
        if not PETSc.Sys.hasExternalPackage('hdf5'):
            return None
        if self.PARTITIONERTYPE != 'simple' and not PETSc.Sys.hasExternalPackage(
            self.PARTITIONERTYPE
        ):
            return None
        return method(self, *args, **kwargs)

    return wrapper


def check_nsize(method):
    def wrapper(self, *args, **kwargs):
        if PETSc.COMM_WORLD.size != self.NSIZE:
            return None
        return method(self, *args, **kwargs)

    return wrapper


class BaseTestPlexHDF5:
    NSIZE = 4
    NTIMES = 3

    def tearDown(self):
        if not PETSc.COMM_WORLD.rank:
            if os.path.exists(self.outfile()):
                os.remove(self.outfile())
            if os.path.exists(self.tmp_output_file()):
                os.remove(self.tmp_output_file())

    def _name(self):
        return f'{self.SUFFIX}_outformat-{self.OUTFORMAT}_{self.PARTITIONERTYPE}'

    def infile(self):
        return os.path.join(
            PETSC_DIR, 'share/petsc/datafiles/', 'meshes/blockcylinder-50.h5'
        )

    def outfile(self):
        return os.path.join('./temp_test_dmplex_%s.h5' % self._name())

    def informat(self):
        return PETSc.Viewer.Format.HDF5_XDMF

    def outformat(self):
        d = {
            'hdf5_petsc': PETSc.Viewer.Format.HDF5_PETSC,
            'hdf5_xdmf': PETSc.Viewer.Format.HDF5_XDMF,
        }
        return d[self.OUTFORMAT]

    def partitionerType(self):
        d = {
            'simple': PETSc.Partitioner.Type.SIMPLE,
            'ptscotch': PETSc.Partitioner.Type.PTSCOTCH,
            'parmetis': PETSc.Partitioner.Type.PARMETIS,
        }
        return d[self.PARTITIONERTYPE]

    def ref_output_file(self):
        return os.path.join(
            PETSC_DIR,
            'src/dm/impls/plex/tutorials/',
            'output/ex5_%s.out' % self._name(),
        )

    def tmp_output_file(self):
        return os.path.join('./temp_test_dmplex_%s.out' % self._name())

    def outputText(self, msg, comm):
        if not comm.rank:
            with open(self.tmp_output_file(), 'a') as f:
                f.write(msg)

    def outputPlex(self, plex):
        txtvwr = PETSc.Viewer().createASCII(
            self.tmp_output_file(), mode='a', comm=plex.comm
        )
        plex.view(viewer=txtvwr)
        txtvwr.destroy()

    @check_dtype
    @check_package
    @check_nsize
    def testViewLoadCycle(self):
        if importlib.util.find_spec('mpi4py') is None:
            self.skipTest('mpi4py')  # throws special exception to signal test skip
        grank = PETSc.COMM_WORLD.rank
        for i in range(self.NTIMES):
            if i == 0:
                infname = self.infile()
                informt = self.informat()
            else:
                infname = self.outfile()
                informt = self.outformat()
            if self.HETEROGENEOUS:
                mycolor = grank > self.NTIMES - i
            else:
                mycolor = 0
            mpicomm = PETSc.COMM_WORLD.tompi4py()
            comm = PETSc.Comm(comm=mpicomm.Split(color=mycolor, key=grank))
            if mycolor == 0:
                self.outputText('Begin cycle %d\n' % i, comm)
                plex = PETSc.DMPlex()
                vwr = PETSc.ViewerHDF5()
                # Create plex
                plex.create(comm=comm)
                plex.setName('DMPlex Object')
                # Load data from XDMF into dm in parallel
                vwr.create(infname, mode='r', comm=comm)
                vwr.pushFormat(format=informt)
                plex.load(viewer=vwr)
                plex.setOptionsPrefix('loaded_')
                plex.distributeSetDefault(False)
                plex.setFromOptions()
                vwr.popFormat()
                vwr.destroy()
                self.outputPlex(plex)
                # Test DM is indeed distributed
                flg = plex.isDistributed()
                self.outputText(
                    'Loaded mesh distributed? %s\n' % str(flg).upper(), comm
                )
                # Interpolate
                plex.interpolate()
                plex.setOptionsPrefix('interpolated_')
                plex.setFromOptions()
                self.outputPlex(plex)
                # Redistribute
                part = plex.getPartitioner()
                part.setType(self.partitionerType())
                sf = plex.distribute(overlap=0)
                if sf:
                    sf.destroy()
                part.destroy()
                plex.setName('DMPlex Object')
                plex.setOptionsPrefix('redistributed_')
                plex.setFromOptions()
                self.outputPlex(plex)
                # Save redistributed dm to XDMF in parallel
                vwr.create(self.outfile(), mode='w', comm=comm)
                vwr.pushFormat(format=self.outformat())
                plex.setName('DMPlex Object')
                plex.view(viewer=vwr)
                vwr.popFormat()
                vwr.destroy()
                # Destroy plex
                plex.destroy()
                self.outputText('End   cycle %d\n--------\n' % i, comm)
            comm.tompi4py().Free()
            PETSc.COMM_WORLD.Barrier()
        # Check that the output is identical to that of plex/tutorial/ex5.c.
        self.assertTrue(
            filecmp.cmp(self.tmp_output_file(), self.ref_output_file(), shallow=False),
            f'Contents of the files not the same. Reference file: {self.ref_output_file()}',
        )
        PETSc.COMM_WORLD.Barrier()


class BaseTestPlexHDF5Homogeneous(BaseTestPlexHDF5):
    """Test save on N / load on N."""

    SUFFIX = 0
    HETEROGENEOUS = False


class BaseTestPlexHDF5Heterogeneous(BaseTestPlexHDF5):
    """Test save on N / load on M."""

    SUFFIX = 1
    HETEROGENEOUS = True


class TestPlexHDF5PETSCSimpleHomogeneous(
    BaseTestPlexHDF5Homogeneous, unittest.TestCase
):
    OUTFORMAT = 'hdf5_petsc'
    PARTITIONERTYPE = 'simple'


"""
Skipping. PTScotch produces different distributions when run
in a sequence in a single session.

class TestPlexHDF5PETSCPTScotchHomogeneous(BaseTestPlexHDF5Homogeneous,
                                           unittest.TestCase):
    OUTFORMAT = "hdf5_petsc"
    PARTITIONERTYPE = "ptscotch"
"""


class TestPlexHDF5PETSCParmetisHomogeneous(
    BaseTestPlexHDF5Homogeneous, unittest.TestCase
):
    OUTFORMAT = 'hdf5_petsc'
    PARTITIONERTYPE = 'parmetis'


class TestPlexHDF5XDMFSimpleHomogeneous(BaseTestPlexHDF5Homogeneous, unittest.TestCase):
    OUTFORMAT = 'hdf5_xdmf'
    PARTITIONERTYPE = 'simple'


"""
Skipping. PTScotch produces different distributions when run
in a sequence in a single session.

class TestPlexHDF5XDMFPTScotchHomogeneous(BaseTestPlexHDF5Homogeneous,
                                          unittest.TestCase):
    OUTFORMAT = "hdf5_xdmf"
    PARTITIONERTYPE = "ptscotch"
"""


class TestPlexHDF5XDMFParmetisHomogeneous(
    BaseTestPlexHDF5Homogeneous, unittest.TestCase
):
    OUTFORMAT = 'hdf5_xdmf'
    PARTITIONERTYPE = 'parmetis'


class TestPlexHDF5PETSCSimpleHeterogeneous(
    BaseTestPlexHDF5Heterogeneous, unittest.TestCase
):
    OUTFORMAT = 'hdf5_petsc'
    PARTITIONERTYPE = 'simple'


"""
Skipping. PTScotch produces different distributions when run
in a sequence in a single session.

class TestPlexHDF5PETSCPTScotchHeterogeneous(BaseTestPlexHDF5Heterogeneous,
                                             unittest.TestCase):
    OUTFORMAT = "hdf5_petsc"
    PARTITIONERTYPE = "ptscotch"
"""


class TestPlexHDF5PETSCParmetisHeterogeneous(
    BaseTestPlexHDF5Heterogeneous, unittest.TestCase
):
    OUTFORMAT = 'hdf5_petsc'
    PARTITIONERTYPE = 'parmetis'


class TestPlexHDF5XDMFSimpleHeterogeneous(
    BaseTestPlexHDF5Heterogeneous, unittest.TestCase
):
    OUTFORMAT = 'hdf5_xdmf'
    PARTITIONERTYPE = 'simple'


class TestPlexHDF5XDMFPTScotchHeterogeneous(
    BaseTestPlexHDF5Heterogeneous, unittest.TestCase
):
    OUTFORMAT = 'hdf5_xdmf'
    PARTITIONERTYPE = 'ptscotch'


class TestPlexHDF5XDMFParmetisHeterogeneous(
    BaseTestPlexHDF5Heterogeneous, unittest.TestCase
):
    OUTFORMAT = 'hdf5_xdmf'
    PARTITIONERTYPE = 'parmetis'


# --------------------------------------------------------------------

if __name__ == '__main__':
    unittest.main()



# --------------------------------------------------------------------

cdef class DMComposite(DM):
    """A DM object that is used to manage data for a collection of DMs."""

    def create(self, comm: Comm | None = None) -> Self:
        """Create a composite object.

        Collective.

        Parameters
        ----------
        comm
            MPI communicator, defaults to `Sys.getDefaultComm`.

        See Also
        --------
        petsc.DMCompositeCreate

        """
        cdef MPI_Comm ccomm = def_Comm(comm, PETSC_COMM_DEFAULT)
        cdef PetscDM newdm = NULL
        CHKERR(DMCompositeCreate(ccomm, &newdm))
        CHKERR(PetscCLEAR(self.obj))
        self.dm = newdm
        return self

    def addDM(self, DM dm, *args: DM) -> None:
        """Add a DM vector to the composite.

        Collective.

        Parameters
        ----------
        dm
            The DM object.
        *args
            Additional DM objects.

        See Also
        --------
        petsc.DMCompositeAddDM

        """
        CHKERR(DMCompositeAddDM(self.dm, dm.dm))
        cdef object item
        for item in args:
            dm = <DM?> item
            CHKERR(DMCompositeAddDM(self.dm, dm.dm))

    def getNumber(self) -> int:
        """Get number of sub-DMs contained in the composite.

        Not collective.

        See Also
        --------
        petsc.DMCompositeGetNumberDM

        """
        cdef PetscInt n = 0
        CHKERR(DMCompositeGetNumberDM(self.dm, &n))
        return toInt(n)
    getNumberDM = getNumber

    def getEntries(self) -> list[DM]:
        """Return sub-DMs contained in the composite.

        Not collective.

        See Also
        --------
        petsc.DMCompositeGetEntriesArray

        """
        cdef PetscInt i, n = 0
        cdef PetscDM *cdms = NULL
        CHKERR(DMCompositeGetNumberDM(self.dm, &n))
        cdef object unused = oarray_p(empty_p(<PetscInt>n), NULL, <void**>&cdms)
        CHKERR(DMCompositeGetEntriesArray(self.dm, cdms))
        cdef DM entry = None
        cdef list entries = []
        for i from 0 <= i < n:
            entry = subtype_DM(cdms[i])()
            entry.dm = cdms[i]
            CHKERR(PetscINCREF(entry.obj))
            entries.append(entry)
        return entries

    def scatter(self, Vec gvec, lvecs: Sequence[Vec]) -> None:
        """Scatter coupled global vector into split local vectors.

        Collective.

        Parameters
        ----------
        gvec
            The global vector.
        lvecs
            Array of local vectors.

        See Also
        --------
        gather, petsc.DMCompositeScatterArray

        """
        cdef PetscInt i, n = 0
        CHKERR(DMCompositeGetNumberDM(self.dm, &n))
        cdef PetscVec *clvecs = NULL
        cdef object unused = oarray_p(empty_p(<PetscInt>n), NULL, <void**>&clvecs)
        for i from 0 <= i < n:
            clvecs[i] = (<Vec?>lvecs[<Py_ssize_t>i]).vec
        CHKERR(DMCompositeScatterArray(self.dm, gvec.vec, clvecs))

    def gather(self, Vec gvec, imode: InsertModeSpec, lvecs: Sequence[Vec]) -> None:
        """Gather split local vectors into a coupled global vector.

        Collective.

        Parameters
        ----------
        gvec
            The global vector.
        imode
            The insertion mode.
        lvecs
            The individual sequential vectors.

        See Also
        --------
        scatter, petsc.DMCompositeGatherArray

        """
        cdef PetscInsertMode cimode = insertmode(imode)
        cdef PetscInt i, n = 0
        CHKERR(DMCompositeGetNumberDM(self.dm, &n))
        cdef PetscVec *clvecs = NULL
        cdef object unused = oarray_p(empty_p(<PetscInt>n), NULL, <void**>&clvecs)
        for i from 0 <= i < n:
            clvecs[i] = (<Vec?>lvecs[<Py_ssize_t>i]).vec
        CHKERR(DMCompositeGatherArray(self.dm, cimode, gvec.vec, clvecs))

    def getGlobalISs(self) -> list[IS]:
        """Return the index sets for each composed object in the composite.

        Collective.

        These could be used to extract a subset of vector entries for a
        "multi-physics" preconditioner.

        Use `getLocalISs` for index sets in the packed local numbering, and
        `getLGMaps` for to map local sub-DM (including ghost) indices to packed
        global indices.

        See Also
        --------
        petsc.DMCompositeGetGlobalISs

        """
        cdef PetscInt i, n = 0
        cdef PetscIS *cis = NULL
        CHKERR(DMCompositeGetNumberDM(self.dm, &n))
        CHKERR(DMCompositeGetGlobalISs(self.dm, &cis))
        cdef object isets = [ref_IS(cis[i]) for i from 0 <= i < n]
        for i from 0 <= i < n:
            CHKERR(ISDestroy(&cis[i]))
        CHKERR(PetscFree(cis))
        return isets

    def getLocalISs(self) -> list[IS]:
        """Return index sets for each component of a composite local vector.

        Not collective.

        To get the composite global indices at all local points (including
        ghosts), use `getLGMaps`.

        To get index sets for pieces of the composite global vector, use
        `getGlobalISs`.

        See Also
        --------
        petsc.DMCompositeGetLocalISs

        """
        cdef PetscInt i, n = 0
        cdef PetscIS *cis = NULL
        CHKERR(DMCompositeGetNumberDM(self.dm, &n))
        CHKERR(DMCompositeGetLocalISs(self.dm, &cis))
        cdef object isets = [ref_IS(cis[i]) for i from 0 <= i < n]
        for i from 0 <= i < n:
            CHKERR(ISDestroy(&cis[i]))
        CHKERR(PetscFree(cis))
        return isets

    def getLGMaps(self) -> list[LGMap]:
        """Return a local-to-global mapping for each DM in the composite.

        Collective.

        Note that this includes all the ghost points that individual ghosted
        DMDA may have.

        See Also
        --------
        petsc.DMCompositeGetISLocalToGlobalMappings

        """
        cdef PetscInt i, n = 0
        cdef PetscLGMap *clgm = NULL
        CHKERR(DMCompositeGetNumberDM(self.dm, &n))
        CHKERR(DMCompositeGetISLocalToGlobalMappings(self.dm, &clgm))
        cdef object lgms = [ref_LGMap(clgm[i]) for i from 0 <= i < n]
        for i from 0 <= i < n:
            CHKERR(ISLocalToGlobalMappingDestroy(&clgm[i]))
        CHKERR(PetscFree(clgm))
        return lgms

    def getAccess(self, Vec gvec, locs: Sequence[int] | None = None) -> Any:
        """Get access to the individual vectors from the global vector.

        Not collective.

        Use via `with` context manager (PEP 343).

        Parameters
        ----------
        gvec
            The global vector.
        locs
            Indices of vectors wanted, or `None` to get all vectors.

        """
        return _DMComposite_access(self, gvec, locs)



# Solves Heat equation on a periodic domain, using raw VecScatter
import sys
import petsc4py

petsc4py.init(sys.argv)

from petsc4py import PETSc
from mpi4py import MPI
import numpy


class Heat:
    def __init__(self, comm, N):
        self.comm = comm
        self.N = N  # global problem size
        self.h = 1 / N  # grid spacing on unit interval
        self.n = N // comm.size + int(
            comm.rank < (N % comm.size)
        )  # owned part of global problem
        self.start = comm.exscan(self.n)
        if comm.rank == 0:
            self.start = 0
        gindices = (
            numpy.arange(self.start - 1, self.start + self.n + 1, dtype=PETSc.IntType)
            % N
        )  # periodic
        self.mat = PETSc.Mat().create(comm=comm)
        size = (self.n, self.N)  # local and global sizes
        self.mat.setSizes((size, size))
        self.mat.setFromOptions()
        self.mat.setPreallocationNNZ(
            (3, 1)
        )  # Conservative preallocation for 3 "local" columns and one non-local

        # Allow matrix insertion using local indices [0:n+2]
        lgmap = PETSc.LGMap().create(list(gindices), comm=comm)
        self.mat.setLGMap(lgmap, lgmap)

        # Global and local vectors
        self.gvec = self.mat.createVecRight()
        self.lvec = PETSc.Vec().create(comm=PETSc.COMM_SELF)
        self.lvec.setSizes(self.n + 2)
        self.lvec.setUp()
        # Configure scatter from global to local
        isg = PETSc.IS().createGeneral(list(gindices), comm=comm)
        self.g2l = PETSc.Scatter().create(self.gvec, isg, self.lvec, None)

        self.tozero, self.zvec = PETSc.Scatter.toZero(self.gvec)
        self.history = []

        if False:  # Print some diagnostics
            print(
                '[%d] local size %d, global size %d, starting offset %d'
                % (comm.rank, self.n, self.N, self.start)
            )
            self.gvec.setArray(numpy.arange(self.start, self.start + self.n))
            self.gvec.view()
            self.g2l.scatter(self.gvec, self.lvec, PETSc.InsertMode.INSERT)
            for rank in range(comm.size):
                if rank == comm.rank:
                    print('Contents of local Vec on rank %d' % rank)
                    self.lvec.view()
                comm.barrier()

    def evalSolution(self, t, x):
        if t != 0.0:
            raise ValueError('Only for t=0')
        coord = numpy.arange(self.start, self.start + self.n) / self.N
        x.setArray((numpy.abs(coord - 0.5) < 0.1) * 1.0)

    def evalFunction(self, ts, t, x, xdot, f):
        self.g2l.scatter(x, self.lvec, PETSc.InsertMode.INSERT)  # lvec is a work vector
        h = self.h
        with self.lvec as u, xdot as udot:
            f.setArray(
                udot * h + 2 * u[1:-1] / h - u[:-2] / h - u[2:] / h
            )  # Scale equation by volume element

    def evalJacobian(self, ts, t, x, xdot, a, A, B):
        h = self.h
        for i in range(self.n):
            lidx = i + 1
            B.setValuesLocal(
                [lidx], [lidx - 1, lidx, lidx + 1], [-1 / h, a * h + 2 / h, -1 / h]
            )
        B.assemble()
        if A != B:
            A.assemble()  # If operator is different from matrix used to construct the preconditioner

    def monitor(self, ts, i, t, x):
        if self.history:
            lasti, lastt, lastx = self.history[-1]
            if i < lasti + 4 or t < lastt + 1e-4:
                return
        self.tozero.scatter(x, self.zvec, PETSc.InsertMode.INSERT)
        xx = self.zvec[:].tolist()
        self.history.append((i, t, xx))

    def plotHistory(self):
        try:
            from matplotlib import pylab, rcParams
        except ImportError:
            return
        rcParams.update({'text.usetex': True, 'figure.figsize': (10, 6)})
        # rc('figure', figsize=(600,400))
        pylab.title('Heat: TS \\texttt{%s}' % ts.getType())
        x = numpy.arange(self.N) / self.N
        for i, t, u in self.history:
            pylab.plot(x, u, label='step=%d t=%8.2g' % (i, t))
        pylab.xlabel('$x$')
        pylab.ylabel('$u$')
        pylab.legend(loc='upper right')
        pylab.savefig('heat-history.png')
        # pylab.show()


OptDB = PETSc.Options()
ode = Heat(MPI.COMM_WORLD, OptDB.getInt('n', 100))

x = ode.gvec.duplicate()
f = ode.gvec.duplicate()

ts = PETSc.TS().create(comm=ode.comm)
ts.setType(ts.Type.ROSW)  # Rosenbrock-W. ARKIMEX is a nonlinearly implicit alternative.

ts.setIFunction(ode.evalFunction, ode.gvec)
ts.setIJacobian(ode.evalJacobian, ode.mat)

ts.setMonitor(ode.monitor)

ts.setTime(0.0)
ts.setTimeStep(ode.h**2)
ts.setMaxTime(1)
ts.setMaxSteps(100)
ts.setExactFinalTime(PETSc.TS.ExactFinalTime.INTERPOLATE)
ts.setMaxSNESFailures(
    -1
)  # allow an unlimited number of failures (step will be rejected and retried)

snes = ts.getSNES()  # Nonlinear solver
snes.setTolerances(
    max_it=10
)  # Stop nonlinear solve after 10 iterations (TS will retry with shorter step)
ksp = snes.getKSP()  # Linear solver
ksp.setType(ksp.Type.CG)  # Conjugate gradients
pc = ksp.getPC()  # Preconditioner
if False:  # Configure algebraic multigrid, could use run-time options instead
    pc.setType(
        pc.Type.GAMG
    )  # PETSc's native AMG implementation, mostly based on smoothed aggregation
    OptDB['mg_coarse_pc_type'] = 'svd'  # more specific multigrid options
    OptDB['mg_levels_pc_type'] = 'sor'

ts.setFromOptions()  # Apply run-time options, e.g. -ts_adapt_monitor -ts_type arkimex -snes_converged_reason
ode.evalSolution(0.0, x)
ts.solve(x)
if ode.comm.rank == 0:
    print(
        'steps %d (%d rejected, %d SNES fails), nonlinear its %d, linear its %d'
        % (
            ts.getStepNumber(),
            ts.getStepRejections(),
            ts.getSNESFailures(),
            ts.getSNESIterations(),
            ts.getKSPIterations(),
        )
    )

if OptDB.getBool('plot_history', True) and ode.comm.rank == 0:
    ode.plotHistory()



# ---------------------------------------------------------------------

cdef inline int setref(void *d, void *s) except -1:
    cdef PetscObject *dest = <PetscObject*> d
    cdef PetscObject source = <PetscObject> s
    CHKERR(PetscINCREF(&source))
    dest[0] = source
    return 0

# ---------------------------------------------------------------------

# -- Error --

cdef api int PyPetscError_Set(PetscErrorCode ierr):
    return SETERR(ierr)

# -- Comm --

cdef api object PyPetscComm_New(MPI_Comm arg):
    cdef Comm retv = Comm()
    retv.comm = arg
    return retv

cdef api MPI_Comm PyPetscComm_Get(object arg) except? MPI_COMM_NULL:
    cdef MPI_Comm retv = MPI_COMM_NULL
    cdef Comm ob = <Comm?> arg
    retv = ob.comm
    return retv

cdef api MPI_Comm* PyPetscComm_GetPtr(object arg) except NULL:
    cdef MPI_Comm *retv = NULL
    cdef Comm ob = <Comm?> arg
    retv = &ob.comm
    return retv

# -- Object --

cdef api object PyPetscObject_New(PetscObject arg):
    cdef Object retv = subtype_Object(arg)()
    setref(&retv.obj[0], arg)
    return retv

cdef api PetscObject PyPetscObject_Get(object arg) except ? NULL:
    cdef PetscObject retv = NULL
    cdef Object ob = <Object?> arg
    retv = ob.obj[0]
    return retv

cdef api PetscObject* PyPetscObject_GetPtr(object arg) except NULL:
    cdef PetscObject *retv = NULL
    cdef Object ob = <Object?> arg
    retv = ob.obj
    return retv

# -- Viewer --

cdef api object PyPetscViewer_New(PetscViewer arg):
    cdef Viewer retv = Viewer()
    setref(&retv.vwr, arg)
    return retv

cdef api PetscViewer PyPetscViewer_Get(object arg) except ? NULL:
    cdef PetscViewer retv = NULL
    cdef Viewer ob = <Viewer?> arg
    retv = ob.vwr
    return retv

# -- Random --

cdef api object PyPetscRandom_New(PetscRandom arg):
    cdef Random retv = Random()
    setref(&retv.rnd, arg)
    return retv

cdef api PetscRandom PyPetscRandom_Get(object arg) except ? NULL:
    cdef PetscRandom retv = NULL
    cdef Random ob = <Random?> arg
    retv = ob.rnd
    return retv

# -- Device --

cdef api Device PyPetscDevice_New(PetscDevice arg):
    cdef Device ret = Device()
    CHKERR(PetscDeviceReference(arg))
    ret.device = arg
    return ret

cdef api PetscDevice PyPetscDevice_Get(object arg) except ? NULL:
    cdef Device      obj = <Device?>arg
    cdef PetscDevice ret = obj.device
    return ret

# -- DeviceContext --

cdef api DeviceContext PyPetscDeviceContext_New(PetscDeviceContext arg):
    cdef DeviceContext retv = DeviceContext()
    setref(&retv.dctx, arg)
    return retv

cdef api PetscDeviceContext PyPetscDeviceContext_Get(object arg) except ? NULL:
    cdef DeviceContext      obj = <DeviceContext?>arg
    cdef PetscDeviceContext ret = obj.dctx
    return ret

# -- IS --

cdef api object PyPetscIS_New(PetscIS arg):
    cdef IS retv = IS()
    setref(&retv.iset, arg)
    return retv

cdef api PetscIS PyPetscIS_Get(object arg) except? NULL:
    cdef PetscIS retv = NULL
    cdef IS ob = <IS?> arg
    retv = ob.iset
    return retv

# -- LGMap --

cdef api object PyPetscLGMap_New(PetscLGMap arg):
    cdef LGMap retv = LGMap()
    setref(&retv.lgm, arg)
    return retv

cdef api PetscLGMap PyPetscLGMap_Get(object arg) except ? NULL:
    cdef PetscLGMap retv = NULL
    cdef LGMap ob = <LGMap?> arg
    retv = ob.lgm
    return retv

# -- SF --

cdef api object PyPetscSF_New(PetscSF arg):
    cdef SF retv = SF()
    setref(&retv.sf, arg)
    return retv

cdef api PetscSF PyPetscSF_Get(object arg) except? NULL:
    cdef PetscSF retv = NULL
    cdef SF ob = <SF?> arg
    retv = ob.sf
    return retv

# -- Vec --

cdef api object PyPetscVec_New(PetscVec arg):
    cdef Vec retv = Vec()
    setref(&retv.vec, arg)
    return retv

cdef api PetscVec PyPetscVec_Get(object arg) except ? NULL:
    cdef PetscVec retv = NULL
    cdef Vec ob = <Vec?> arg
    retv = ob.vec
    return retv

# -- Scatter --

cdef api object PyPetscScatter_New(PetscScatter arg):
    cdef Scatter retv = Scatter()
    setref(&retv.sct, arg)
    return retv

cdef api PetscScatter PyPetscScatter_Get(object arg) except ? NULL:
    cdef PetscScatter retv = NULL
    cdef Scatter ob = <Scatter?> arg
    retv = ob.sct
    return retv

# -- Section --

cdef api object PyPetscSection_New(PetscSection arg):
    cdef Section retv = Section()
    setref(&retv.sec, arg)
    return retv

cdef api PetscSection PyPetscSection_Get(object arg) except ? NULL:
    cdef PetscSection retv = NULL
    cdef Section ob = <Section?> arg
    retv = ob.sec
    return retv

# -- Mat --

cdef api object PyPetscMat_New(PetscMat arg):
    cdef Mat retv = Mat()
    setref(&retv.mat, arg)
    return retv

cdef api PetscMat PyPetscMat_Get(object arg) except ? NULL:
    cdef PetscMat retv = NULL
    cdef Mat ob = <Mat?> arg
    retv = ob.mat
    return retv

# -- MatPartitioning --

cdef api object PyPetscMatPartitioning_New(PetscMatPartitioning arg):
    cdef MatPartitioning retv = MatPartitioning()
    setref(&retv.part, arg)
    return retv

cdef api PetscMatPartitioning PyPetscMatPartitioning_Get(object arg) except ? NULL:
    cdef PetscMatPartitioning retv = NULL
    cdef MatPartitioning ob = <MatPartitioning?> arg
    retv = ob.part
    return retv

# -- PC --

cdef api object PyPetscPC_New(PetscPC arg):
    cdef PC retv = PC()
    setref(&retv.pc, arg)
    return retv

cdef api PetscPC PyPetscPC_Get(object arg) except ? NULL:
    cdef PetscPC retv = NULL
    cdef PC ob = <PC?> arg
    retv = ob.pc
    return retv

# -- KSP --

cdef api object PyPetscKSP_New(PetscKSP arg):
    cdef KSP retv = KSP()
    setref(&retv.ksp, arg)
    return retv

cdef api PetscKSP PyPetscKSP_Get(object arg) except ? NULL:
    cdef PetscKSP retv = NULL
    cdef KSP ob = <KSP?> arg
    retv = ob.ksp
    return retv

# -- SNES --

cdef api object PyPetscSNES_New(PetscSNES arg):
    cdef SNES retv = SNES()
    setref(&retv.snes, arg)
    return retv

cdef api PetscSNES PyPetscSNES_Get(object arg) except ? NULL:
    cdef PetscSNES retv = NULL
    cdef SNES ob = <SNES?> arg
    retv = ob.snes
    return retv

cdef api object PyPetscSNESLineSearch_New(PetscSNESLineSearch arg):
    cdef SNESLineSearch retv = SNESLineSearch()
    setref(&retv.snesls, arg)
    return retv

cdef api PetscSNESLineSearch PyPetscSNESLineSearch_Get(object arg) except ? NULL:
    cdef PetscSNESLineSearch retv = NULL
    cdef SNESLineSearch ob = <SNESLineSearch?> arg
    retv = ob.snesls
    return retv

# -- TS --

cdef api object PyPetscTS_New(PetscTS arg):
    cdef TS retv = TS()
    setref(&retv.ts, arg)
    return retv

cdef api PetscTS PyPetscTS_Get(object arg) except ? NULL:
    cdef PetscTS retv = NULL
    cdef TS ob = <TS?> arg
    retv = ob.ts
    return retv

# -- TAO --

cdef api object PyPetscTAO_New(PetscTAO arg):
    cdef TAO retv = TAO()
    setref(&retv.tao, arg)
    return retv

cdef api PetscTAO PyPetscTAO_Get(object arg) except ? NULL:
    cdef PetscTAO retv = NULL
    cdef TAO ob = <TAO?> arg
    retv = ob.tao
    return retv

cdef api object PyPetscTAOLineSearch_New(PetscTAOLineSearch arg):
    cdef TAOLineSearch retv = TAOLineSearch()
    setref(&retv.taols, arg)
    return retv

cdef api PetscTAOLineSearch PyPetscTAOLineSearch_Get(object arg) except ? NULL:
    cdef PetscTAOLineSearch retv = NULL
    cdef TAOLineSearch ob = <TAOLineSearch?> arg
    retv = ob.taols
    return retv


# -- AO --

cdef api object PyPetscAO_New(PetscAO arg):
    cdef AO retv = AO()
    setref(&retv.ao, arg)
    return retv

cdef api PetscAO PyPetscAO_Get(object arg) except ? NULL:
    cdef PetscAO retv = NULL
    cdef AO ob = <AO?> arg
    retv = ob.ao
    return retv

# -- DM --

cdef api object PyPetscDM_New(PetscDM arg):
    cdef DM retv = subtype_DM(arg)()
    setref(&retv.dm, arg)
    return retv

cdef api PetscDM PyPetscDM_Get(object arg) except ? NULL:
    cdef PetscDM retv = NULL
    cdef DM ob = <DM?> arg
    retv = ob.dm
    return retv

# -- DS --

cdef api object PyPetscDS_New(PetscDS arg):
    cdef DS retv = DS()
    setref(&retv.ds, arg)
    return retv

cdef api PetscDS PyPetscDS_Get(object arg) except ? NULL:
    cdef PetscDS retv = NULL
    cdef DS ob = <DS?> arg
    retv = ob.ds
    return retv

# -- Partitioner --

cdef api object PyPetscPartitioner_New(PetscPartitioner arg):
    cdef Partitioner retv = Partitioner()
    setref(&retv.part, arg)
    return retv

cdef api PetscPartitioner PyPetscPartitioner_Get(object arg) except ? NULL:
    cdef PetscPartitioner retv = NULL
    cdef Partitioner ob = <Partitioner?> arg
    retv = ob.part
    return retv

# -- FE --

cdef api object PyPetscFE_New(PetscFE arg):
    cdef FE retv = FE()
    setref(&retv.fe, arg)
    return retv

cdef api PetscFE PyPetscFE_Get(object arg) except ? NULL:
    cdef PetscFE retv = NULL
    cdef FE ob = <FE?> arg
    retv = ob.fe
    return retv

# -- QUAD --

cdef api object PyPetscQuad_New(PetscQuadrature arg):
    cdef Quad retv = Quad()
    setref(&retv.quad, arg)
    return retv

cdef api PetscQuadrature PyPetscQuad_Get(object arg) except ? NULL:
    cdef PetscQuadrature retv = NULL
    cdef Quad ob = <Quad?> arg
    retv = ob.quad
    return retv

# -- DMLabel --

cdef api object PyPetscDMLabel_New(PetscDMLabel arg):
    cdef DMLabel retv = DMLabel()
    setref(&retv.dmlabel, arg)
    return retv

cdef api PetscDMLabel PyPetscDMLabel_Get(object arg) except ? NULL:
    cdef PetscDMLabel retv = NULL
    cdef DMLabel ob = <DMLabel?> arg
    retv = ob.dmlabel
    return retv


# -- SPACE --

cdef api object PyPetscSpace_New(PetscSpace arg):
    cdef Space retv = Space()
    setref(&retv.space, arg)
    return retv

cdef api PetscSpace PyPetscSpace_Get(object arg) except ? NULL:
    cdef PetscSpace retv = NULL
    cdef Space ob = <Space?> arg
    retv = ob.space
    return retv

# -- DUALSPACE --

cdef api object PyPetscDualSpace_New(PetscDualSpace arg):
    cdef DualSpace retv = DualSpace()
    setref(&retv.dualspace, arg)
    return retv

cdef api PetscDualSpace PyPetscDualSpace_Get(object arg) except ? NULL:
    cdef PetscDualSpace retv = NULL
    cdef DualSpace ob = <DualSpace?> arg
    retv = ob.dualspace
    return retv

# -- DMSwarmCellDM --

cdef api object PyPetscDMSwarmCellDM_New(PetscDMSwarmCellDM arg):
    cdef CellDM retv = CellDM()
    setref(&retv.cdm, arg)
    return retv

cdef api PetscDMSwarmCellDM PyPetscDMSwarmCellDM_Get(object arg) except ? NULL:
    cdef PetscDMSwarmCellDM retv = NULL
    cdef CellDM ob = <CellDM?> arg
    retv = ob.cdm
    return retv

# -- PetscRegressor --

cdef api object PyPetscRegressor_New(PetscRegressor arg):
    cdef Regressor retv = Regressor()
    setref(&retv.regressor, arg)
    return retv

cdef api PetscRegressor PyPetscRegressor_Get(object arg) except ? NULL:
    cdef PetscRegressor retv = NULL
    cdef Regressor ob = <Regressor?> arg
    retv = ob.regressor
    return retv

# ---------------------------------------------------------------------



import pyvista as pv
import numpy as np
from petsc4py import PETSc

SCALAR = 0
VECTOR = 1

def _convertCell(ctype, cells, nc, off):
    # The VTK conventions are at https://www.princeton.edu/~efeibush/viscourse/vtk.pdf
    if ctype == PETSc.DM.PolytopeType.TETRAHEDRON:
        tmp = cells[off + 1]
        cells[off + 1] = cells[off + 2]
        cells[off + 2] = tmp
    elif ctype == PETSc.DM.PolytopeType.HEXAHEDRON:
        tmp = cells[off + 1]
        cells[off + 1] = cells[off + 3]
        cells[off + 3] = tmp
    elif ctype == PETSc.DM.PolytopeType.TRI_PRISM:
        tmp = cells[off + 4]
        cells[off + 4] = cells[off + 5]
        cells[off + 5] = tmp
    elif ctype == PETSc.DM.PolytopeType.TRI_PRISM_TENSOR:
        tmp = cells[off + 1]
        cells[off + 1] = cells[off + 2]
        cells[off + 2] = tmp
        tmp = cells[off + 4]
        cells[off + 4] = cells[off + 5]
        cells[off + 5] = tmp
    elif ctype == PETSc.DM.PolytopeType.PYRAMID:
        tmp            = cells[off + 1]
        cells[off + 1] = cells[off + 3]
        cells[off + 3] = tmp
    return


VTK_TYPES = {}
VTK_TYPES[PETSc.DM.PolytopeType.POINT] = pv.CellType.VERTEX
VTK_TYPES[PETSc.DM.PolytopeType.SEGMENT] = pv.CellType.LINE
VTK_TYPES[PETSc.DM.PolytopeType.TRIANGLE] = pv.CellType.TRIANGLE
VTK_TYPES[PETSc.DM.PolytopeType.QUADRILATERAL] = pv.CellType.QUAD
VTK_TYPES[PETSc.DM.PolytopeType.TETRAHEDRON] = pv.CellType.TETRA
VTK_TYPES[PETSc.DM.PolytopeType.HEXAHEDRON] = pv.CellType.HEXAHEDRON
VTK_TYPES[PETSc.DM.PolytopeType.TRI_PRISM] = pv.CellType.WEDGE
VTK_TYPES[PETSc.DM.PolytopeType.TRI_PRISM_TENSOR] = pv.CellType.WEDGE
VTK_TYPES[PETSc.DM.PolytopeType.QUAD_PRISM_TENSOR] = pv.CellType.HEXAHEDRON
VTK_TYPES[PETSc.DM.PolytopeType.PYRAMID] = pv.CellType.PYRAMID


class PetscPyVista:
    def setUp(self, viewer):
        pass

    def setFromOptions(self, viewer):
        OptDB = PETSc.Options(viewer.prefix)
        self.swarmField     = OptDB.getString('view_pyvista_swarm_field', 'w_q')
        self.swarmPointSize = OptDB.getInt('view_pyvista_swarm_point_size', 5)
        self.warpFactor     = OptDB.getReal('view_pyvista_warp', 0.0)
        self.clipBounds     = OptDB.getRealArray('view_pyvista_clip', [])
        self.glyphScale     = OptDB.getReal('view_pyvista_glyph_scale', 0.0)

    def view(self, viewer, outviewer):
        pass

    def flush(self, viewer):
        pass

    def convertDMToPV(self, plex):
        cdim = plex.getCoordinateDim()
        vStart, vEnd = plex.getDepthStratum(0)
        cStart, cEnd = plex.getHeightStratum(0)
        conesLength = 0
        # Maybe it will be faster in C?
        # DMPlexGetCellsVertices?
        for c in range(cStart, cEnd):
            conesLength += 1
            closure, ornt = plex.getTransitiveClosure(c)
            for cl in closure:
                if cl >= vStart and cl < vEnd:
                    conesLength += 1
        cells = np.zeros((conesLength), dtype=np.uint32)
        conesLength = 0
        for c in range(cStart, cEnd):
            closure, ornt = plex.getTransitiveClosure(c)
            nc = 0
            off = 1
            for cl in closure:
                if cl >= vStart and cl < vEnd:
                    cells[conesLength] += 1
                    cells[conesLength + off] = cl - vStart
                    nc += 1
                    off += 1
            _convertCell(plex.getCellType(c), cells, nc, conesLength + 1)
            conesLength += off
        celltypes = np.zeros((cEnd - cStart), dtype=np.uint32)
        for c in range(cStart, cEnd):
            celltypes[c] = VTK_TYPES[plex.getCellType(c)]
        points = np.zeros((vEnd - vStart, 3), dtype=np.float32)
        with plex.getCoordinatesLocal().getBuffer() as coords:
            for v in range(vEnd - vStart):
                for d in range(cdim):
                    points[v, d] = coords[v * cdim + d]
        return pv.UnstructuredGrid(cells, celltypes, points)

    def viewPlex(self, viewer, dm, scalars = None):
        grid  = self.convertDMToPV(dm)
        name  = viewer.getFileName()
        ftype = None
        dim   = dm.getDimension()
        if scalars is not None:
            if scalars[2] == 1:
                ftype = SCALAR
            elif scalars[2] == dim:
                ftype = VECTOR
            else:
                raise RuntimeError('Scalars \'%s\' blocksize %d did not match 1 or mesh dim %d' % (scalars[0], scalars[2], dm.getDimension()))
            if scalars[1].shape[0] / scalars[2] == grid.n_cells:
                grid.cell_data[scalars[0]] = scalars[1]
            elif scalars[1].shape[0] / scalars[2] == grid.n_points:
                if dim == 3:
                    grid.point_data[scalars[0]] = scalars[1].reshape(-1, scalars[2])
                else:
                    vecs = np.zeros((scalars[1].shape[0] // scalars[2], 3))
                    vecs[:, 0:2] = scalars[1].reshape(-1, scalars[2])
                    grid.point_data[scalars[0]] = vecs
            else:
                raise RuntimeError('Scalars \'%s\' size %d (%d) did not match sizes for cells (%d) or vertices (%d)' % (scalars[0], scalars[1].shape[0], scalars[2], grid.n_cells, grid.n_points))
            if self.warpFactor > 0.:
                if ftype == SCALAR:
                    grid = grid.warp_by_scalar(factor = self.warpFactor)
                elif ftype == VECTOR:
                    grid = grid.warp_by_vector(factor = self.warpFactor)
        if len(self.clipBounds) == 6:
            grid = grid.clip_box(self.clipBounds)
        elif len(self.clipBounds) == 3:
            grid = grid.clip(self.clipBounds)
        if name is None:
            pl = pv.Plotter()
            if ftype == VECTOR:
                pl.add_mesh(grid, show_edges=True)
                if self.glyphScale > 0.:
                    grid.point_data["magnitudes"] = self.glyphScale * np.linalg.norm(grid.point_data[scalars[0]], axis=1)
                    pl.add_mesh(grid.glyph(orient=scalars[0], scale="magnitudes"))
                else:
                    pl.add_mesh(grid.glyph(orient=scalars[0], scale=scalars[0]))
            elif ftype == SCALAR:
                pl.add_mesh(grid, show_edges=True, scalars=scalars[0])
            else:
                pl.add_mesh(grid, show_edges=True)
            pl.show()
        else:
            grid.plot(show_edges=True, scalars=scalars[0], off_screen=True, screenshot=name)
        return

    def viewSwarm(self, viewer, sw):
        import math
        name = viewer.getFileName()
        spoints = sw.getField('DMSwarmPIC_coor')
        n       = spoints.shape[0]
        bs      = spoints.shape[1]
        points  = np.zeros((n, 3))
        for i in range(n):
            points[i,:bs] = spoints[i,:]
        vpoints = sw.getField('velocity')
        nv      = vpoints.shape[0]
        vbs     = vpoints.shape[1]
        vpoints = vpoints.reshape((nv, vbs))
        wgt   = sw.getField(self.swarmField)
        field = np.zeros((n,))
        for i in range(n):
            field[i] = wgt[i, 0]
        if name is None:
            pl = pv.Plotter()
        else:
            pl = pv.Plotter(off_screen=True)
        pl.add_points(points, scalars=field, render_points_as_spheres=False, point_size=self.swarmPointSize, name="swarm")
        maxF = field.max()
        for i in range(n):
            if maxF <= 0.:
              continue
            if math.fabs(field[i]) < 0.1 * maxF:
              continue
            if math.fabs(vpoints[i,0]) > 0.001:
              continue
            val = 2. * math.fabs(field[i]) / maxF
            pl.add_mesh(pv.Disc(center = points[i,:], normal = (1., 0., 0.), inner = 0.25 * val, outer = val), opacity = 0.2)
        if name is None:
            pl.show()
        else:
            pl.show(interactive=False,screenshot=name)
        sw.restoreField(self.swarmField)
        sw.restoreField('velocity')
        sw.restoreField('DMSwarmPIC_coor')
        return

    def viewObject(self, viewer, pobj):
        if pobj.klass == 'Vec':
          dm = pobj.getDM()
          a = pobj.getArray(readonly=1)
          bs = pobj.getBlockSize()
          self.viewPlex(viewer, dm, scalars = (pobj.name, a, bs))
        elif pobj.klass == 'DM':
            if pobj.type == 'plex':
                self.viewPlex(viewer, pobj)
            elif pobj.type == 'swarm':
                self.viewSwarm(viewer, pobj)
        return

    def viewCell(self, grid, c):
        cell = grid.get_cell(c)
        print(cell)
        cell.plot(show_edges=True)
        return



# --------------------------------------------------------------------

cdef extern from * nogil:

    ctypedef enum PetscDMSwarmType "DMSwarmType":
        DMSWARM_BASIC
        DMSWARM_PIC

    ctypedef enum PetscDMSwarmMigrateType "DMSwarmMigrateType":
        DMSWARM_MIGRATE_BASIC
        DMSWARM_MIGRATE_DMCELLNSCATTER
        DMSWARM_MIGRATE_DMCELLEXACT
        DMSWARM_MIGRATE_USER

    ctypedef enum PetscDMSwarmCollectType "DMSwarmCollectType":
        DMSWARM_COLLECT_BASIC
        DMSWARM_COLLECT_DMDABOUNDINGBOX
        DMSWARM_COLLECT_GENERAL
        DMSWARM_COLLECT_USER

    ctypedef enum PetscDMSwarmPICLayoutType 'DMSwarmPICLayoutType':
        DMSWARMPIC_LAYOUT_REGULAR
        DMSWARMPIC_LAYOUT_GAUSS
        DMSWARMPIC_LAYOUT_SUBDIVISION

    PetscErrorCode DMSwarmCreateGlobalVectorFromField(PetscDM, const char[], PetscVec*)
    PetscErrorCode DMSwarmDestroyGlobalVectorFromField(PetscDM, const char[], PetscVec*)
    PetscErrorCode DMSwarmCreateLocalVectorFromField(PetscDM, const char[], PetscVec*)
    PetscErrorCode DMSwarmDestroyLocalVectorFromField(PetscDM, const char[], PetscVec*)
    PetscErrorCode DMSwarmCreateGlobalVectorFromFields(PetscDM, PetscInt, const char *[], PetscVec *)
    PetscErrorCode DMSwarmDestroyGlobalVectorFromFields(PetscDM, PetscInt, const char *[], PetscVec *)
    PetscErrorCode DMSwarmCreateLocalVectorFromFields(PetscDM, PetscInt, const char *[], PetscVec *)
    PetscErrorCode DMSwarmDestroyLocalVectorFromFields(PetscDM, PetscInt, const char *[], PetscVec *)

    PetscErrorCode DMSwarmInitializeFieldRegister(PetscDM)
    PetscErrorCode DMSwarmFinalizeFieldRegister(PetscDM)
    PetscErrorCode DMSwarmSetLocalSizes(PetscDM, PetscInt, PetscInt)
    PetscErrorCode DMSwarmRegisterPetscDatatypeField(PetscDM, const char[], PetscInt, PetscDataType)
#    PetscErrorCode DMSwarmRegisterUserStructField(PetscDM, const char[], size_t)
#    PetscErrorCode DMSwarmRegisterUserDatatypeField(PetscDM, const char[], size_t, PetscInt)
    PetscErrorCode DMSwarmGetField(PetscDM, const char[], PetscInt*, PetscDataType*, void**)
    PetscErrorCode DMSwarmRestoreField(PetscDM, const char[], PetscInt*, PetscDataType*, void**)
    PetscErrorCode DMSwarmGetFieldInfo(PetscDM, const char[], PetscInt *, PetscDataType *)

    PetscErrorCode DMSwarmVectorDefineField(PetscDM, const char[])

    PetscErrorCode DMSwarmAddPoint(PetscDM)
    PetscErrorCode DMSwarmAddNPoints(PetscDM, PetscInt)
    PetscErrorCode DMSwarmRemovePoint(PetscDM)
    PetscErrorCode DMSwarmRemovePointAtIndex(PetscDM, PetscInt)
    PetscErrorCode DMSwarmCopyPoint(PetscDM, PetscInt, PetscInt)

    PetscErrorCode DMSwarmGetLocalSize(PetscDM, PetscInt*)
    PetscErrorCode DMSwarmGetSize(PetscDM, PetscInt*)
    PetscErrorCode DMSwarmMigrate(PetscDM, PetscBool)

    PetscErrorCode DMSwarmCollectViewCreate(PetscDM)
    PetscErrorCode DMSwarmCollectViewDestroy(PetscDM)
    PetscErrorCode DMSwarmSetCellDM(PetscDM, PetscDM)
    PetscErrorCode DMSwarmGetCellDM(PetscDM, PetscDM*)
    PetscErrorCode DMSwarmGetCellDMByName(PetscDM, const char[], PetscDMSwarmCellDM *)
    PetscErrorCode DMSwarmGetCellDMNames(PetscDM, PetscInt *, const char **[])
    PetscErrorCode DMSwarmSetCellDMActive(PetscDM, const char[])
    PetscErrorCode DMSwarmGetCellDMActive(PetscDM, PetscDMSwarmCellDM *)
    PetscErrorCode DMSwarmAddCellDM(PetscDM, PetscDMSwarmCellDM)

    PetscErrorCode DMSwarmSetType(PetscDM, PetscDMSwarmType)

    PetscErrorCode DMSwarmSetPointsUniformCoordinates(PetscDM, PetscReal[], PetscReal[], PetscInt[], PetscInsertMode)
    PetscErrorCode DMSwarmSetPointCoordinates(PetscDM, PetscInt, PetscReal*, PetscBool, PetscInsertMode)
    PetscErrorCode DMSwarmInsertPointsUsingCellDM(PetscDM, PetscDMSwarmPICLayoutType, PetscInt)
    PetscErrorCode DMSwarmSetPointCoordinatesCellwise(PetscDM, PetscInt, PetscReal*)
    PetscErrorCode DMSwarmViewFieldsXDMF(PetscDM, const char*, PetscInt, const char**)
    PetscErrorCode DMSwarmViewXDMF(PetscDM, const char*)

    PetscErrorCode DMSwarmSortGetAccess(PetscDM)
    PetscErrorCode DMSwarmSortRestoreAccess(PetscDM)
    PetscErrorCode DMSwarmSortGetPointsPerCell(PetscDM, PetscInt, PetscInt*, PetscInt**)
    PetscErrorCode DMSwarmSortGetNumberOfPointsPerCell(PetscDM, PetscInt, PetscInt*)
    PetscErrorCode DMSwarmSortGetIsValid(PetscDM, PetscBool*)
    PetscErrorCode DMSwarmSortGetSizes(PetscDM, PetscInt*, PetscInt*)

    PetscErrorCode DMSwarmProjectFields(PetscDM, PetscDM, PetscInt, const char**, PetscVec*, PetscScatterMode)

    PetscErrorCode DMSwarmComputeMoments(PetscDM, const char[], const char[], PetscReal[])

    PetscErrorCode DMSwarmCellDMCreate(PetscDM, PetscInt, const char *[], PetscInt, const char *[], PetscDMSwarmCellDM *)
    PetscErrorCode DMSwarmCellDMDestroy(PetscDMSwarmCellDM *)
    PetscErrorCode DMSwarmCellDMView(PetscDMSwarmCellDM, PetscViewer)
    PetscErrorCode DMSwarmCellDMGetDM(PetscDMSwarmCellDM, PetscDM *)
    PetscErrorCode DMSwarmCellDMGetFields(PetscDMSwarmCellDM, PetscInt *, const char **[])
    PetscErrorCode DMSwarmCellDMGetCoordinateFields(PetscDMSwarmCellDM, PetscInt *, const char **[])
    PetscErrorCode DMSwarmCellDMGetCellID(PetscDMSwarmCellDM, const char *[])
    PetscErrorCode DMSwarmCellDMGetBlockSize(PetscDMSwarmCellDM, PetscDM, PetscInt *)



cdef extern from * nogil:

    enum: PETSC_DECIDE
    enum: PETSC_DEFAULT
    enum: PETSC_DETERMINE
    enum: PETSC_CURRENT
    enum: PETSC_UNLIMITED

    PetscReal PETSC_INFINITY
    PetscReal PETSC_NINFINITY

    ctypedef enum PetscInsertMode "InsertMode":
        PETSC_NOT_SET_VALUES    "NOT_SET_VALUES"
        PETSC_INSERT_VALUES     "INSERT_VALUES"
        PETSC_ADD_VALUES        "ADD_VALUES"
        PETSC_MAX_VALUES        "MAX_VALUES"
        PETSC_INSERT_ALL_VALUES "INSERT_ALL_VALUES"
        PETSC_ADD_ALL_VALUES    "ADD_ALL_VALUES"
        PETSC_INSERT_BC_VALUES  "INSERT_BC_VALUES"
        PETSC_ADD_BC_VALUES     "ADD_BC_VALUES"

    ctypedef enum PetscScatterMode "ScatterMode":
        PETSC_SCATTER_FORWARD       "SCATTER_FORWARD"
        PETSC_SCATTER_REVERSE       "SCATTER_REVERSE"
        PETSC_SCATTER_FORWARD_LOCAL "SCATTER_FORWARD_LOCAL"
        PETSC_SCATTER_REVERSE_LOCAL "SCATTER_REVERSE_LOCAL"

    ctypedef enum  PetscNormType "NormType":
        PETSC_NORM_1          "NORM_1"
        PETSC_NORM_2          "NORM_2"
        PETSC_NORM_1_AND_2    "NORM_1_AND_2"
        PETSC_NORM_FROBENIUS  "NORM_FROBENIUS"
        PETSC_NORM_INFINITY   "NORM_INFINITY"
        PETSC_NORM_MAX        "NORM_MAX"

    ctypedef enum PetscCopyMode:
        PETSC_COPY_VALUES
        PETSC_OWN_POINTER
        PETSC_USE_POINTER

cdef inline PetscInsertMode insertmode(object mode) \
    except <PetscInsertMode>(-1):
    if   mode is None:  return PETSC_INSERT_VALUES
    elif mode is True:  return PETSC_ADD_VALUES
    elif mode is False: return PETSC_INSERT_VALUES
    else:               return mode

cdef inline PetscScatterMode scattermode(object mode) \
    except <PetscScatterMode>(-1):
    if mode is None:  return PETSC_SCATTER_FORWARD
    if mode is False: return PETSC_SCATTER_FORWARD
    if mode is True:  return PETSC_SCATTER_REVERSE
    if isinstance(mode, str):
        if mode == 'forward': return PETSC_SCATTER_FORWARD
        if mode == 'reverse': return PETSC_SCATTER_REVERSE
        else: raise ValueError("unknown scatter mode: %s" % mode)
    return mode



cdef extern from * nogil:

    ctypedef const char* PetscDSType
    PetscDSType PETSCDSBASIC

    PetscErrorCode PetscDSCreate(MPI_Comm, PetscDS*)
    PetscErrorCode PetscDSDestroy(PetscDS*)
    PetscErrorCode PetscDSView(PetscDS, PetscViewer)
    PetscErrorCode PetscDSSetType(PetscDS, PetscDSType)
    PetscErrorCode PetscDSGetType(PetscDS, PetscDSType*)
    PetscErrorCode PetscDSSetFromOptions(PetscDS)
    PetscErrorCode PetscDSSetUp(PetscDS)

    PetscErrorCode PetscDSGetHeightSubspace(PetscDS, PetscInt, PetscDS*)
    PetscErrorCode PetscDSGetSpatialDimension(PetscDS, PetscInt*)
    PetscErrorCode PetscDSGetCoordinateDimension(PetscDS, PetscInt*)
    PetscErrorCode PetscDSSetCoordinateDimension(PetscDS, PetscInt)
    PetscErrorCode PetscDSGetNumFields(PetscDS, PetscInt*)
    PetscErrorCode PetscDSGetTotalDimension(PetscDS, PetscInt*)
    PetscErrorCode PetscDSGetTotalComponents(PetscDS, PetscInt*)
    PetscErrorCode PetscDSGetFieldIndex(PetscDS, PetscObject, PetscInt*)
    PetscErrorCode PetscDSGetFieldSize(PetscDS, PetscInt, PetscInt*)
    PetscErrorCode PetscDSGetFieldOffset(PetscDS, PetscInt, PetscInt*)
    PetscErrorCode PetscDSGetDimensions(PetscDS, PetscInt*[])
    PetscErrorCode PetscDSGetComponents(PetscDS, PetscInt*[])
    PetscErrorCode PetscDSGetComponentOffset(PetscDS, PetscInt, PetscInt*)
    PetscErrorCode PetscDSGetComponentOffsets(PetscDS, PetscInt*[])
    PetscErrorCode PetscDSGetComponentDerivativeOffsets(PetscDS, PetscInt*[])

    PetscErrorCode PetscDSGetDiscretization(PetscDS, PetscInt, PetscObject*)
    PetscErrorCode PetscDSSetDiscretization(PetscDS, PetscInt, PetscObject)
    PetscErrorCode PetscDSAddDiscretization(PetscDS, PetscObject)
    PetscErrorCode PetscDSGetImplicit(PetscDS, PetscInt, PetscBool*)
    PetscErrorCode PetscDSSetImplicit(PetscDS, PetscInt, PetscBool)



#!/usr/bin/env python3
# Author:  Lisandro Dalcin
# Contact: dalcinl@gmail.com

import re
import os
import sys
import warnings

try:
    import setuptools
except ImportError:
    setuptools = None

topdir = os.path.abspath(os.path.dirname(__file__))
sys.path.insert(0, os.path.join(topdir, 'conf'))

pyver = sys.version_info[:2]
if pyver < (3, 6):
    raise RuntimeError('Python version 3.6 or higher is required')

PNAME = 'PETSc'
EMAIL = 'petsc-maint@mcs.anl.gov'
PLIST = [PNAME]

# --------------------------------------------------------------------
# Metadata
# --------------------------------------------------------------------

py_limited_api = (3, 10)


def F(string):
    return string.format(
        Name=PNAME,
        name=PNAME.lower(),
        pyname=PNAME.lower() + '4py',
    )


def get_name():
    return F('{pyname}')


def get_version():
    try:
        return get_version.result
    except AttributeError:
        pass
    pkg_init_py = os.path.join(F('{pyname}'), '__init__.py')
    with open(os.path.join(topdir, 'src', pkg_init_py)) as f:
        m = re.search(r"__version__\s*=\s*'(.*)'", f.read())
    version = m.groups()[0]
    get_version.result = version
    return version


def description():
    return F('{Name} for Python')


def long_description():
    with open(os.path.join(topdir, 'DESCRIPTION.rst')) as f:
        return f.read()


url = F('https://gitlab.com/{name}/{name}')
pypiroot = F('https://pypi.io/packages/source')
pypislug = F('{pyname}')[0] + F('/{pyname}')
tarball = F('{pyname}-%s.tar.gz' % get_version())
download = '/'.join([pypiroot, pypislug, tarball])

classifiers = """
Operating System :: POSIX
Intended Audience :: Developers
Intended Audience :: Science/Research
Programming Language :: C
Programming Language :: C++
Programming Language :: Cython
Programming Language :: Python
Programming Language :: Python :: 3
Programming Language :: Python :: Implementation :: CPython
Topic :: Scientific/Engineering
Topic :: Software Development :: Libraries :: Python Modules
Development Status :: 5 - Production/Stable
""".strip().split('\n')

keywords = """
scientific computing
parallel computing
MPI
""".strip().split('\n')

platforms = """
POSIX
Linux
macOS
FreeBSD
""".strip().split('\n')

metadata = {
    'name': get_name(),
    'version': get_version(),
    'description': description(),
    'long_description': long_description(),
    'url': url,
    'download_url': download,
    'classifiers': classifiers,
    'keywords': keywords + PLIST,
    'license': 'BSD-2-Clause',
    'platforms': platforms,
    'author': 'Lisandro Dalcin',
    'author_email': 'dalcinl@gmail.com',
    'maintainer': F('{Name} Team'),
    'maintainer_email': EMAIL,
}
metadata.update(
    {
        'requires': ['numpy'],
    }
)

metadata_extra = {
    'long_description_content_type': 'text/x-rst',
}

def get_build_pysabi():
    abi = os.environ.get("PETSC4PY_BUILD_PYSABI", "").lower()
    if abi and sys.implementation.name == "cpython":
        if abi in {"false", "no", "off", "n", "0"}:
            return None
        if abi in {"true", "yes", "on", "y", "1"} | {"abi3"}:
            return py_limited_api
        if abi.startswith("cp"):
            abi = abi[2:]
        if "." in abi:
            x, y = abi.split(".")
        else:
            x, y = abi[0], abi[1:]
        return (int(x), int(y))
    return None

# --------------------------------------------------------------------
# Extension modules
# --------------------------------------------------------------------


def sources():
    src = {
        'source': F('{pyname}/{Name}.pyx'),
        'depends': [
            F('{pyname}/*.pyx'),
            F('{pyname}/*.pxd'),
            F('{pyname}/{Name}/*.pyx'),
            F('{pyname}/{Name}/*.pxd'),
            F('{pyname}/{Name}/*.pxi'),
        ],
        'workdir': 'src',
    }
    return [src]


def extensions():
    from os import walk
    from glob import glob
    from os.path import join

    #
    depends = []
    glob_join = lambda *args: glob(join(*args))
    for pth, _, _ in walk('src'):
        depends += glob_join(pth, '*.h')
        depends += glob_join(pth, '*.c')
    for pkg in map(str.lower, reversed(PLIST)):
        if (pkg.upper() + '_DIR') in os.environ:
            pd = os.environ[pkg.upper() + '_DIR']
            pa = os.environ.get('PETSC_ARCH', '')
            depends += glob_join(pd, 'include', '*.h')
            depends += glob_join(pd, 'include', pkg, 'private', '*.h')
            depends += glob_join(pd, pa, 'include', '%sconf.h' % pkg)
    #
    include_dirs = []
    numpy_include = os.environ.get('NUMPY_INCLUDE')
    if numpy_include is not None:
        numpy_includes = [numpy_include]
    else:
        try:
            import numpy

            numpy_includes = [numpy.get_include()]
        except ImportError:
            numpy_includes = []
    include_dirs.extend(numpy_includes)
    if F('{pyname}') != 'petsc4py':
        try:
            import petsc4py

            petsc4py_includes = [petsc4py.get_include()]
        except ImportError:
            petsc4py_includes = []
        include_dirs.extend(petsc4py_includes)
    #
    ext = {
        'name': F('{pyname}.lib.{Name}'),
        'sources': [F('src/{pyname}/{Name}.c')],
        'depends': depends,
        'include_dirs': [
            'src',
            F('src/{pyname}/include'),
        ]
        + include_dirs,
        'define_macros': [
            ('MPICH_SKIP_MPICXX', 1),
            ('OMPI_SKIP_MPICXX', 1),
            ('NPY_NO_DEPRECATED_API', 'NPY_1_7_API_VERSION'),
        ],
    }
    return [ext]


# --------------------------------------------------------------------
# Setup
# --------------------------------------------------------------------


def get_release():
    suffix = os.path.join('src', 'binding', F('{pyname}'))
    if not topdir.endswith(os.path.join(os.path.sep, suffix)):
        return True
    release = 1
    rootdir = os.path.abspath(os.path.join(topdir, *[os.path.pardir] * 3))
    version_h = os.path.join(rootdir, 'include', F('{name}version.h'))
    release_macro = '%s_VERSION_RELEASE' % F('{name}').upper()
    version_re = re.compile(r'#define\s+%s\s+([-]*\d+)' % release_macro)
    if os.path.exists(version_h) and os.path.isfile(version_h):
        with open(version_h, 'r') as f:
            release = int(version_re.search(f.read()).groups()[0])
    return bool(release)


def requires(pkgname, major, minor, release=True):
    minor = minor + int(not release)
    devel = '' if release else '.dev0'
    vmin = f'{major}.{minor}{devel}'
    vmax = f'{major}.{minor+1}'
    return f'{pkgname}>={vmin},<{vmax}'


def run_setup():
    is_sdist = 'sdist' in sys.argv
    setup_args = metadata.copy()
    vstr = setup_args['version'].split('.')[:2]
    x, y = tuple(map(int, vstr))
    release = get_release()
    if not release:
        setup_args['version'] = '%d.%d.0.dev0' % (x, y + 1)
    if setuptools:
        warnings.filterwarnings(
            'ignore', message=r'.*fetch_build_eggs', module='setuptools'
        )
        setup_args['zip_safe'] = False
        numpy_pin = 'numpy'
        if not is_sdist:
            try:
                import numpy

                major = int(numpy.__version__.partition('.')[0])
                numpy_pin = 'numpy>=1.19' if major >= 2 else 'numpy<2'
            except ImportError:
                pass
        setup_args['setup_requires'] = ['numpy']
        setup_args['install_requires'] = [numpy_pin]
        for pkg in map(str.lower, PLIST):
            PKG_DIR = os.environ.get(pkg.upper() + '_DIR')
            if not (PKG_DIR and os.path.isdir(PKG_DIR)):
                package = requires(pkg, x, y, release)
                setup_args['setup_requires'] += [package]
                setup_args['install_requires'] += [package]
        if F('{pyname}') != 'petsc4py':
            package = requires('petsc4py', x, y, release)
            setup_args['setup_requires'] += [package]
            setup_args['install_requires'] += [package]
        setup_args.update(metadata_extra)
    #
    conf = __import__(F('conf{name}'))
    cython_sources = [src for src in sources()]  # noqa: C416
    ext_modules = [conf.Extension(**ext) for ext in extensions()]
    #
    sabi = get_build_pysabi()
    if sabi and setuptools:
        api_tag = "cp{}{}".format(*sabi)
        options = {"bdist_wheel": {"py_limited_api": api_tag}}
        setup_args["options"] = options
        api_ver = "0x{:02X}{:02X}0000".format(*sabi)
        defines = [("Py_LIMITED_API", api_ver)]
        for ext in ext_modules:
            ext.define_macros.extend(defines)
            ext.py_limited_api = True
    #
    conf.setup(
        packages=[
            F('{pyname}'),
            F('{pyname}.lib'),
            F('{pyname}.lib._pytypes'),
            F('{pyname}.lib._pytypes.viewer'),
        ],
        package_dir={'': 'src'},
        package_data={
            F('{pyname}'): [
                F('{Name}.pxd'),
                F('{Name}*.h'),
                F('include/{pyname}/*.h'),
                F('include/{pyname}/*.i'),
                'py.typed',
                '*.pyi',
                '*/*.pyi',
            ],
            F('{pyname}.lib'): [
                F('{name}.cfg'),
            ],
        },
        cython_sources=cython_sources,
        ext_modules=ext_modules,
        **setup_args,
    )


# --------------------------------------------------------------------


def main():
    run_setup()


if __name__ == '__main__':
    main()

# --------------------------------------------------------------------



# Linear regression with synthetic data generation
# ================================================
#
# Generate a synthetic data set using sklearn.datasets.make_regression() and
# then solve it using the petsc4py interface to PetscRegressor.

import numpy as np
# Needed for plotting
import matplotlib.colors
import matplotlib.pyplot as plt

# Needed for generating classification, regression and clustering datasets
import sklearn.datasets as dt

# Needed for evaluating the quality of the regression for the test data set
from sklearn.metrics import mean_squared_error

import argparse
parser = argparse.ArgumentParser('Test MLG using synthetic data')
parser.add_argument('--nsample', type=int, default=10000)
parser.add_argument('--nfeature', type=int, default=10)
parser.add_argument('--add_noise', action='store_true')
args, unknown = parser.parse_known_args()

import sys
import petsc4py
sys.argv = [sys.argv[0]] + unknown
petsc4py.init(sys.argv)
from petsc4py import PETSc

# Define the seed so that results can be reproduced
seed = 11
rand_state = 11

# Define the color maps for plots
color_map = plt.cm.get_cmap('RdYlBu')
color_map_discrete = matplotlib.colors.LinearSegmentedColormap.from_list("", ["red","cyan","magenta","blue"])

def petsc_regression_test(nsample, nfeature, noise=0, rand_state=11):
  ntr = round(nsample*9/10)
  nte = nsample - ntr
  x,y = dt.make_regression(n_samples=nsample,
                           n_features=nfeature,
                           noise=noise,
                           random_state=rand_state)
  x_train, y_train = x[:ntr,], y[:ntr]
  xte, yte = x[ntr:,], y[ntr:]
  comm = PETSc.COMM_WORLD
  rank = comm.getRank()
  regressor = PETSc.Regressor().create(comm=comm)
  regressor.setType(PETSc.Regressor.Type.LINEAR)
  rows_ix = np.arange(ntr,dtype=np.int32)
  cols_ix = np.arange(nfeature,dtype=np.int32)
  X = PETSc.Mat().create(comm=comm)
  X.setSizes((ntr,nfeature))
  X.setFromOptions()
  X.setUp()
  y = PETSc.Vec().create(comm=comm)
  y.setSizes(ntr)
  y.setFromOptions()
  if not rank :
    X.setValues(rows_ix,cols_ix,x_train,addv=True)
    y.setValues(rows_ix,y_train,addv=False)
  X.assemblyBegin(X.AssemblyType.FINAL)
  X.assemblyEnd(X.AssemblyType.FINAL)
  y.assemblyBegin()
  y.assemblyEnd()
  regressor.fit(X,y)
  rows_ix = np.arange(nte,dtype=np.int32)
  X = PETSc.Mat().create(comm=comm)
  X.setSizes((nte,nfeature))
  X.setFromOptions()
  X.setUp()
  y = PETSc.Vec().create(comm=comm)
  y.setSizes(nte)
  y.setFromOptions()
  if not rank :
    X.setValues(rows_ix,cols_ix,xte,addv=True)
    y.zeroEntries()
  X.assemblyBegin(X.AssemblyType.FINAL)
  X.assemblyEnd(X.AssemblyType.FINAL)
  y.assemblyBegin()
  y.assemblyEnd()
  regressor.predict(X,y)
  ypr = y.getArray()
  error = mean_squared_error(ypr,yte)
  print(f"Test MSE: {error:f}")
  return xte,ypr

xte,ypr = petsc_regression_test(args.nsample, args.nfeature, 0)

if args.add_noise:
  fig,ax = plt.subplots(nrows=2, ncols=3,figsize=(16,7))
  plt_ind_list = np.arange(6)+231

  for noise,plt_ind in zip([0,0.1,1,10,100,1000],plt_ind_list):
    xte,ypr = petsc_regression_test(args.nsample, args.nfeature, noise, rand_state)

    plt.subplot(plt_ind)
    my_scatter_plot = plt.scatter(xte[:,0],
                                  xte[:,1],
                                  c=ypr,
                                  vmin=min(ypr),
                                  vmax=max(ypr),
                                  s=35,
                                  cmap=color_map)

    plt.title('noise: '+str(noise))
    plt.colorbar(my_scatter_plot)

  fig.subplots_adjust(hspace=0.3,wspace=.3)
  plt.suptitle('PETSc regression tests with different noise levels',fontsize=20)
  plt.show(block=True)



import sys,petsc4py
petsc4py.init(sys.argv)
from petsc4py import PETSc
import numpy as np


def sensor(x, y):
    """
    Classic hyperbolic sensor function for testing
    multi-scale anisotropic mesh adaptation:

    f:[-1, 1]² → R,
        f(x, y) = sin(50xy)/100 if |xy| > 2π/50 else sin(50xy)

    (mapped to have domain [0,1]² in this case).
    """
    xy = (2*x - 1)*(2*y - 1)
    ret = np.sin(50*xy)
    if np.abs(xy) > 2*np.pi/50:
        ret *= 0.01
    return ret


# Set metric parameters
h_min = 1.0e-10             # Minimum tolerated metric magnitude ~ cell size
h_max = 1.0e-01             # Maximum tolerated metric magnitude ~ cell size
a_max = 1.0e+05             # Maximum tolerated anisotropy
targetComplexity = 10000.0  # Analogous to number of vertices in adapted mesh
p = 1.0                     # Lᵖ normalization order

# Create a uniform mesh
OptDB = PETSc.Options()
dim = OptDB.getInt('dim', 2)
numEdges = 10
simplex = True
plex = PETSc.DMPlex().createBoxMesh([numEdges]*dim, simplex=simplex)
plex.distribute()
plex.view()
viewer = PETSc.Viewer().createVTK("anisotropic_mesh_0.vtk", "w")
viewer(plex)

# Do four mesh adaptation iterations
for i in range(1, 5):
    vStart, vEnd = plex.getDepthStratum(0)

    # Create a P1 sensor function
    comm = plex.getComm()
    fe = PETSc.FE().createLagrange(dim, 1, simplex, 1, -1, comm=comm)
    plex.setField(0, fe)
    plex.createDS()
    f = plex.createLocalVector()
    csec = plex.getCoordinateSection()
    coords = plex.getCoordinatesLocal()
    pf = f.getArray()
    pcoords = coords.getArray()
    for v in range(vStart, vEnd):
        off = csec.getOffset(v)
        x = pcoords[off]
        y = pcoords[off+1]
        pf[off//dim] = sensor(x, y)
    f.setName("Sensor")
    viewer = PETSc.Viewer().createVTK(f"sensor_{i}.vtk", "w")
    viewer(f)

    # Recover the gradient of the sensor function
    dmGrad = plex.clone()
    fe = PETSc.FE().createLagrange(dim, dim, simplex, 1, -1, comm=comm)
    dmGrad.setField(0, fe)
    dmGrad.createDS()
    g = dmGrad.createLocalVector()
    plex.computeGradientClementInterpolant(f, g)
    g.setName("Gradient")
    viewer = PETSc.Viewer().createVTK(f"gradient_{i}.vtk", "w")
    viewer(g)

    # Recover the Hessian of the sensor function
    dmHess = plex.clone()
    H = dmHess.metricCreate()
    dmGrad.computeGradientClementInterpolant(g, H)
    H.setName("Hessian")
    viewer = PETSc.Viewer().createVTK(f"hessian_{i}.vtk", "w")
    viewer(H)

    # Obtain a metric by Lᵖ normalization
    dmHess.metricSetMinimumMagnitude(h_min)
    dmHess.metricSetMaximumMagnitude(h_max)
    dmHess.metricSetMaximumAnisotropy(a_max)
    dmHess.metricSetNormalizationOrder(p)
    dmHess.metricSetTargetComplexity(targetComplexity)
    metric = dmHess.metricCreate()
    det = dmHess.metricDeterminantCreate()
    dmHess.metricNormalize(H, metric, det)
    metric.setName("Metric")
    viewer = PETSc.Viewer().createVTK(f"metric_{i}.vtk", "w")
    viewer(metric)

    # Call adapt routine - boundary label None by default
    plex = plex.adaptMetric(metric)
    plex.distribute()
    plex.view()

    # Write to VTK file
    viewer = PETSc.Viewer().createVTK(f"anisotropic_mesh_{i}.vtk", "w")
    viewer(plex)



# --------------------------------------------------------------------

class ViewerType(object):
    """Viewer type."""
    SOCKET      = S_(PETSCVIEWERSOCKET)
    ASCII       = S_(PETSCVIEWERASCII)
    BINARY      = S_(PETSCVIEWERBINARY)
    STRING      = S_(PETSCVIEWERSTRING)
    DRAW        = S_(PETSCVIEWERDRAW)
    VU          = S_(PETSCVIEWERVU)
    MATHEMATICA = S_(PETSCVIEWERMATHEMATICA)
    HDF5        = S_(PETSCVIEWERHDF5)
    VTK         = S_(PETSCVIEWERVTK)
    MATLAB      = S_(PETSCVIEWERMATLAB)
    SAWS        = S_(PETSCVIEWERSAWS)
    GLVIS       = S_(PETSCVIEWERGLVIS)
    ADIOS       = S_(PETSCVIEWERADIOS)
    EXODUSII    = S_(PETSCVIEWEREXODUSII)
    PYTHON      = S_(PETSCVIEWERPYTHON)
    PYVISTA     = S_(PETSCVIEWERPYVISTA)


class ViewerFormat(object):
    """Viewer format."""
    DEFAULT           = PETSC_VIEWER_DEFAULT
    ASCII_MATLAB      = PETSC_VIEWER_ASCII_MATLAB
    ASCII_MATHEMATICA = PETSC_VIEWER_ASCII_MATHEMATICA
    ASCII_IMPL        = PETSC_VIEWER_ASCII_IMPL
    ASCII_INFO        = PETSC_VIEWER_ASCII_INFO
    ASCII_INFO_DETAIL = PETSC_VIEWER_ASCII_INFO_DETAIL
    ASCII_COMMON      = PETSC_VIEWER_ASCII_COMMON
    ASCII_SYMMODU     = PETSC_VIEWER_ASCII_SYMMODU
    ASCII_INDEX       = PETSC_VIEWER_ASCII_INDEX
    ASCII_DENSE       = PETSC_VIEWER_ASCII_DENSE
    ASCII_MATRIXMARKET= PETSC_VIEWER_ASCII_MATRIXMARKET
    ASCII_PCICE       = PETSC_VIEWER_ASCII_PCICE
    ASCII_PYTHON      = PETSC_VIEWER_ASCII_PYTHON
    ASCII_FACTOR_INFO = PETSC_VIEWER_ASCII_FACTOR_INFO
    ASCII_LATEX       = PETSC_VIEWER_ASCII_LATEX
    ASCII_XML         = PETSC_VIEWER_ASCII_XML
    ASCII_GLVIS       = PETSC_VIEWER_ASCII_GLVIS
    ASCII_CSV         = PETSC_VIEWER_ASCII_CSV
    DRAW_BASIC        = PETSC_VIEWER_DRAW_BASIC
    DRAW_LG           = PETSC_VIEWER_DRAW_LG
    DRAW_LG_XRANGE    = PETSC_VIEWER_DRAW_LG_XRANGE
    DRAW_CONTOUR      = PETSC_VIEWER_DRAW_CONTOUR
    DRAW_PORTS        = PETSC_VIEWER_DRAW_PORTS
    VTK_VTS           = PETSC_VIEWER_VTK_VTS
    VTK_VTR           = PETSC_VIEWER_VTK_VTR
    VTK_VTU           = PETSC_VIEWER_VTK_VTU
    BINARY_MATLAB     = PETSC_VIEWER_BINARY_MATLAB
    NATIVE            = PETSC_VIEWER_NATIVE
    HDF5_PETSC        = PETSC_VIEWER_HDF5_PETSC
    HDF5_VIZ          = PETSC_VIEWER_HDF5_VIZ
    HDF5_XDMF         = PETSC_VIEWER_HDF5_XDMF
    HDF5_MAT          = PETSC_VIEWER_HDF5_MAT
    NOFORMAT          = PETSC_VIEWER_NOFORMAT
    LOAD_BALANCE      = PETSC_VIEWER_LOAD_BALANCE
    FAILED            = PETSC_VIEWER_FAILED


class ViewerFileMode(object):
    """Viewer file mode."""
    # native
    READ          = PETSC_FILE_MODE_READ
    WRITE         = PETSC_FILE_MODE_WRITE
    APPEND        = PETSC_FILE_MODE_APPEND
    UPDATE        = PETSC_FILE_MODE_UPDATE
    APPEND_UPDATE = PETSC_FILE_MODE_APPEND_UPDATE
    # aliases
    R, W, A, U = READ, WRITE, APPEND, UPDATE
    AU = UA    = APPEND_UPDATE


class ViewerDrawSize(object):
    """Window size."""
    # native
    FULL_SIZE    = PETSC_DRAW_FULL_SIZE
    HALF_SIZE    = PETSC_DRAW_HALF_SIZE
    THIRD_SIZE   = PETSC_DRAW_THIRD_SIZE
    QUARTER_SIZE = PETSC_DRAW_QUARTER_SIZE
    # aliases
    FULL    = FULL_SIZE
    HALF    = HALF_SIZE
    THIRD   = THIRD_SIZE
    QUARTER = QUARTER_SIZE

# --------------------------------------------------------------------


cdef class Viewer(Object):
    """Viewer object.

    Viewer is described in the `PETSc manual <petsc:sec_viewers>`.

    Viewers can be called as functions where the argument specified
    is the PETSc object to be viewed. See the example below.

    Examples
    --------
    >>> from petsc4py import PETSc
    >>> u = PETSc.Vec().createWithArray([1,2])
    >>> v = PETSc.Viewer()
    >>> v(u)
    Vec Object: 1 MPI process
      type: seq
    1.
    2.

    See Also
    --------
    petsc.PetscViewer

    """

    Type   = ViewerType
    Format = ViewerFormat
    FileMode = ViewerFileMode
    DrawSize = ViewerDrawSize

    # backward compatibility
    Mode = ViewerFileMode
    Size = ViewerFileMode

    #

    def __cinit__(self):
        self.obj = <PetscObject*> &self.vwr
        self.vwr = NULL

    def __call__(self, Object obj) -> None:
        """View a generic object."""
        assert obj.obj != NULL
        CHKERR(PetscObjectView(obj.obj[0], self.vwr))

    #

    def view(self, obj: Viewer | Object | None = None) -> None:
        """View the viewer.

        Collective.

        Parameters
        ----------
        obj
            A `Viewer` instance or `None` for the default viewer.
            If none of the above applies, it assumes ``obj`` is an instance of `Object`
            and it calls the generic view for ``obj``.

        Notes
        -----

        See Also
        --------
        petsc.PetscViewerView

        """
        if obj is None:
            CHKERR(PetscViewerView(self.vwr, NULL))
        elif isinstance(obj, Viewer):
            CHKERR(PetscViewerView(self.vwr, (<Viewer?>obj).vwr))
        else:
            assert (<Object?>obj).obj != NULL
            CHKERR(PetscObjectView((<Object?>obj).obj[0], self.vwr))

    def destroy(self) -> Self:
        """Destroy the viewer.

        Collective.

        See Also
        --------
        petsc.PetscViewerDestroy

        """
        CHKERR(PetscViewerDestroy(&self.vwr))
        return self

    def create(self, comm: Comm | None = None) -> Self:
        """Create a viewer.

        Collective.

        Parameters
        ----------
        comm
            MPI communicator, defaults to `Sys.getDefaultComm`.

        See Also
        --------
        Sys.getDefaultComm, petsc.PetscViewerCreate

        """
        cdef MPI_Comm ccomm = def_Comm(comm, PETSC_COMM_DEFAULT)
        cdef PetscViewer newvwr = NULL
        CHKERR(PetscViewerCreate(ccomm, &newvwr))
        CHKERR(PetscCLEAR(self.obj)); self.vwr = newvwr
        return self

    def createASCII(
        self,
        name: str,
        mode: FileMode | str | None = None,
        comm: Comm | None = None,
       ) -> Self:
        """Create a viewer of type `Type.ASCII`.

        Collective.

        Parameters
        ----------
        name
            The filename associated with the viewer.
        mode
            The mode type.
        comm
            MPI communicator, defaults to `Sys.getDefaultComm`.

        See Also
        --------
        create, setType, setFileMode, setFileName, Sys.getDefaultComm
        setASCIITab, addASCIITab, subtractASCIITab, getASCIITab

        """
        cdef MPI_Comm ccomm = def_Comm(comm, PETSC_COMM_DEFAULT)
        cdef const char *cname = NULL
        name = str2bytes(name, &cname)
        cdef PetscFileMode cmode = PETSC_FILE_MODE_WRITE
        if mode is not None: cmode = filemode(mode)
        cdef PetscViewer newvwr = NULL
        CHKERR(PetscViewerCreate(ccomm, &newvwr))
        CHKERR(PetscCLEAR(self.obj)); self.vwr = newvwr
        CHKERR(PetscViewerSetType(self.vwr, PETSCVIEWERASCII))
        CHKERR(PetscViewerFileSetMode(self.vwr, cmode))
        CHKERR(PetscViewerFileSetName(self.vwr, cname))
        return self

    def createBinary(
        self,
        name: str,
        mode: FileMode | str | None = None,
        comm: Comm | None = None,
       ) -> Self:
        """Create a viewer of type `Type.BINARY`.

        Collective.

        Parameters
        ----------
        name
            The filename associated with the viewer.
        mode
            The mode type.
        comm
            MPI communicator, defaults to `Sys.getDefaultComm`.

        See Also
        --------
        create, setType, setFileMode, setFileName, Sys.getDefaultComm

        """
        cdef MPI_Comm ccomm = def_Comm(comm, PETSC_COMM_DEFAULT)
        cdef const char *cname = NULL
        name = str2bytes(name, &cname)
        cdef PetscFileMode cmode = filemode(mode)
        cdef PetscViewer newvwr = NULL
        CHKERR(PetscViewerBinaryOpen(ccomm, cname, cmode, &newvwr))
        CHKERR(PetscCLEAR(self.obj)); self.vwr = newvwr
        return self

    def createMPIIO(
        self,
        name: str,
        mode: FileMode | str | None = None,
        comm: Comm | None = None,
       ) -> Self:
        """Create a viewer of type `Type.BINARY` supporting MPI-IO.

        Collective.

        Parameters
        ----------
        name
            The filename associated with the viewer.
        mode
            The mode type.
        comm
            MPI communicator, defaults to `Sys.getDefaultComm`.

        See Also
        --------
        create, setType, setFileMode, setFileName, Sys.getDefaultComm

        """
        cdef MPI_Comm ccomm = def_Comm(comm, PETSC_COMM_DEFAULT)
        cdef const char *cname = NULL
        name = str2bytes(name, &cname)
        cdef PetscFileMode cmode = filemode(mode)
        cdef PetscViewer newvwr = NULL
        CHKERR(PetscViewerCreate(ccomm, &newvwr))
        CHKERR(PetscCLEAR(self.obj)); self.vwr = newvwr
        CHKERR(PetscViewerSetType(self.vwr, PETSCVIEWERBINARY))
        CHKERR(PetscViewerBinarySetUseMPIIO(self.vwr, PETSC_TRUE))
        CHKERR(PetscViewerFileSetMode(self.vwr, cmode))
        CHKERR(PetscViewerFileSetName(self.vwr, cname))
        return self

    def createVTK(
        self,
        name: str,
        mode: FileMode | str | None = None,
        comm: Comm | None = None,
       ) -> Self:
        """Create a viewer of type `Type.VTK`.

        Collective.

        Parameters
        ----------
        name
            The filename associated with the viewer.
        mode
            The mode type.
        comm
            MPI communicator, defaults to `Sys.getDefaultComm`.

        See Also
        --------
        create, setType, setFileMode, setFileName, Sys.getDefaultComm

        """
        cdef MPI_Comm ccomm = def_Comm(comm, PETSC_COMM_DEFAULT)
        cdef const char *cname = NULL
        name = str2bytes(name, &cname)
        cdef PetscFileMode cmode = filemode(mode)
        cdef PetscViewer newvwr = NULL
        CHKERR(PetscViewerCreate(ccomm, &newvwr))
        CHKERR(PetscCLEAR(self.obj)); self.vwr = newvwr
        CHKERR(PetscViewerSetType(self.vwr, PETSCVIEWERVTK))
        CHKERR(PetscViewerFileSetMode(self.vwr, cmode))
        CHKERR(PetscViewerFileSetName(self.vwr, cname))
        return self

    def createHDF5(
        self,
        name: str,
        mode: FileMode | str | None = None,
        comm: Comm | None = None,
       ) -> Self:
        """Create a viewer of type `Type.HDF5`.

        Collective.

        Parameters
        ----------
        name
            The filename associated with the viewer.
        mode
            The mode type.
        comm
            MPI communicator, defaults to `Sys.getDefaultComm`.

        See Also
        --------
        create, setType, setFileMode, setFileName, Sys.getDefaultComm

        """
        cdef MPI_Comm ccomm = def_Comm(comm, PETSC_COMM_DEFAULT)
        cdef const char *cname = NULL
        name = str2bytes(name, &cname)
        cdef PetscFileMode cmode = filemode(mode)
        cdef PetscViewer newvwr = NULL
        CHKERR(PetscViewerCreate(ccomm, &newvwr))
        CHKERR(PetscCLEAR(self.obj)); self.vwr = newvwr
        CHKERR(PetscViewerSetType(self.vwr, PETSCVIEWERHDF5))
        CHKERR(PetscViewerFileSetMode(self.vwr, cmode))
        CHKERR(PetscViewerFileSetName(self.vwr, cname))
        return self

    def createDraw(
        self,
        display: str | None = None,
        title: str | None = None,
        position: tuple[int, int] | None = None,
        size: tuple[int, int] | int | None = None,
        comm: Comm | None = None,
       ) -> Self:
        """Create a `Type.DRAW` viewer.

        Collective.

        Parameters
        ----------
        display
            The X display to use or `None` for the local machine.
        title
            The window title or `None` for no title.
        position
            Screen coordinates of the upper left corner, or `None` for default.
        size
            Window size or `None` for default.
        comm
            MPI communicator, defaults to `Sys.getDefaultComm`.

        See Also
        --------
        Sys.getDefaultComm, petsc.PetscViewerDrawOpen

        """
        cdef MPI_Comm ccomm = def_Comm(comm, PETSC_COMM_DEFAULT)
        cdef const char *cdisplay = NULL
        cdef const char *ctitle = NULL
        display = str2bytes(display, &cdisplay)
        title = str2bytes(title, &ctitle)
        cdef int x, y, h, w
        x = y = h = w = PETSC_DECIDE
        if position not in (None, PETSC_DECIDE):
            x, y = position
        if size not in (None, PETSC_DECIDE):
            try:
                w, h = size
            except TypeError:
                w = h = size
        cdef PetscViewer newvwr = NULL
        CHKERR(PetscViewerDrawOpen(ccomm, cdisplay, ctitle,
                                   x, y, w, h, &newvwr))
        CHKERR(PetscCLEAR(self.obj)); self.vwr = newvwr
        return self

    def setType(self, vwr_type: Type | str) -> None:
        """Set the type of the viewer.

        Logically collective.

        Parameters
        ----------
        vwr_type
            The type of the viewer.

        See Also
        --------
        getType, petsc.PetscViewerSetType

        """
        cdef PetscViewerType cval = NULL
        vwr_type = str2bytes(vwr_type, &cval)
        CHKERR(PetscViewerSetType(self.vwr, cval))

    def getType(self) -> str:
        """Return the type of the viewer.

        Not collective.

        See Also
        --------
        setType, petsc.PetscViewerGetType

        """
        cdef PetscViewerType cval = NULL
        CHKERR(PetscViewerGetType(self.vwr, &cval))
        return bytes2str(cval)

    def setFromOptions(self) -> None:
        """Configure the object from the options database.

        Collective.

        See Also
        --------
        petsc_options, petsc.PetscViewerSetFromOptions

        """
        CHKERR(PetscViewerSetFromOptions(self.vwr))

    def setUp(self) -> Self:
        """Set up the internal data structures for using the viewer.

        Collective.

        See Also
        --------
        petsc.PetscViewerSetUp

        """
        CHKERR(PetscViewerSetUp(self.vwr))
        return self

    def getFormat(self) -> Format:
        """Return the format of the viewer.

        Not collective.

        See Also
        --------
        pushFormat, popFormat, petsc.PetscViewerGetFormat

        """
        cdef PetscViewerFormat format = PETSC_VIEWER_DEFAULT
        CHKERR(PetscViewerGetFormat(self.vwr, &format))
        return format

    def pushFormat(self, format: Format) -> None:
        """Push format to the viewer.

        Collective.

        See Also
        --------
        popFormat, petsc.PetscViewerPushFormat

        """
        CHKERR(PetscViewerPushFormat(self.vwr, format))

    def popFormat(self) -> None:
        """Pop format from the viewer.

        Collective.

        See Also
        --------
        pushFormat, petsc.PetscViewerPopFormat

        """
        CHKERR(PetscViewerPopFormat(self.vwr))

    def getSubViewer(self, comm: Comm | None = None) -> Viewer:
        """Return a viewer defined on a subcommunicator.

        Collective.

        Parameters
        ----------
        comm
            The subcommunicator. If `None`, uses `COMM_SELF`.

        Notes
        -----
        Users must call `restoreSubViewer` when done.

        See Also
        --------
        restoreSubViewer, petsc.PetscViewerGetSubViewer

        """
        cdef MPI_Comm ccomm = def_Comm(comm, PETSC_COMM_SELF)
        cdef Viewer sub = Viewer()
        CHKERR(PetscViewerGetSubViewer(self.vwr, ccomm, &sub.vwr))
        return sub

    def restoreSubViewer(self, Viewer sub) -> None:
        """Restore a viewer defined on a subcommunicator.

        Collective.

        Parameters
        ----------
        sub
            The subviewer obtained from `getSubViewer`.

        See Also
        --------
        getSubViewer, petsc.PetscViewerRestoreSubViewer

        """
        cdef MPI_Comm ccomm = def_Comm(sub.getComm(), PETSC_COMM_SELF)
        CHKERR(PetscViewerRestoreSubViewer(self.vwr, ccomm, &sub.vwr))

    @classmethod
    def STDOUT(cls, comm: Comm | None = None) -> Viewer:
        """Return the standard output viewer associated with the communicator.

        Collective.

        Parameters
        ----------
        comm
            MPI communicator, defaults to `Sys.getDefaultComm`.

        """
        cdef MPI_Comm ccomm = def_Comm(comm, PETSC_COMM_DEFAULT)
        cdef Viewer viewer = Viewer()
        viewer.vwr = PETSC_VIEWER_STDOUT_(ccomm)
        CHKERR(PetscINCREF(viewer.obj))
        return viewer

    @classmethod
    def STDERR(cls, comm: Comm | None = None) -> Viewer:
        """Return the standard error viewer associated with the communicator.

        Collective.

        Parameters
        ----------
        comm
            MPI communicator, defaults to `Sys.getDefaultComm`.

        """
        cdef MPI_Comm ccomm = def_Comm(comm, PETSC_COMM_DEFAULT)
        cdef Viewer viewer = Viewer()
        viewer.vwr = PETSC_VIEWER_STDERR_(ccomm)
        CHKERR(PetscINCREF(viewer.obj))
        return viewer

    @classmethod
    def ASCII(cls, name : str, comm: Comm | None = None) -> Viewer:
        """Return an ASCII viewer associated with the communicator.

        Collective.

        Parameters
        ----------
        name
            The filename.
        comm
            MPI communicator, defaults to `Sys.getDefaultComm`.

        """
        cdef MPI_Comm ccomm = def_Comm(comm, PETSC_COMM_DEFAULT)
        cdef const char *cname = NULL
        name = str2bytes(name, &cname)
        cdef Viewer viewer = Viewer()
        CHKERR(PetscViewerASCIIOpen(ccomm, cname, &viewer.vwr))
        return viewer

    @classmethod
    def BINARY(cls, comm: Comm | None = None) -> Viewer:
        """Return the default `Type.BINARY` viewer associated with the communicator.

        Collective.

        Parameters
        ----------
        comm
            MPI communicator, defaults to `Sys.getDefaultComm`.

        """
        cdef MPI_Comm ccomm = def_Comm(comm, PETSC_COMM_DEFAULT)
        cdef Viewer viewer = Viewer()
        viewer.vwr = PETSC_VIEWER_BINARY_(ccomm)
        CHKERR(PetscINCREF(viewer.obj))
        return viewer

    @classmethod
    def DRAW(cls, comm: Comm | None = None) -> Viewer:
        """Return the default `Type.DRAW` viewer associated with the communicator.

        Collective.

        Parameters
        ----------
        comm
            MPI communicator, defaults to `Sys.getDefaultComm`.

        """
        cdef MPI_Comm ccomm = def_Comm(comm, PETSC_COMM_DEFAULT)
        cdef Viewer viewer = Viewer()
        viewer.vwr = PETSC_VIEWER_DRAW_(ccomm)
        CHKERR(PetscINCREF(viewer.obj))
        return viewer

    # --- ASCII viewers ---

    def setASCIITab(self, tabs : int) -> None:
        """Set ASCII tab level.

        Collective.

        See Also
        --------
        getASCIITab, petsc.PetscViewerASCIISetTab

        """
        cdef PetscInt ctabs = asInt(tabs)
        CHKERR(PetscViewerASCIISetTab(self.vwr, ctabs))

    def getASCIITab(self) -> int:
        """Return the ASCII tab level.

        Not collective.

        See Also
        --------
        setASCIITab, petsc.PetscViewerASCIIGetTab

        """
        cdef PetscInt tabs = 0
        CHKERR(PetscViewerASCIIGetTab(self.vwr, &tabs))
        return toInt(tabs)

    def addASCIITab(self, tabs: int) -> None:
        """Increment the ASCII tab level.

        Collective.

        See Also
        --------
        petsc.PetscViewerASCIIAddTab

        """
        cdef PetscInt ctabs = asInt(tabs)
        CHKERR(PetscViewerASCIIAddTab(self.vwr, ctabs))

    def subtractASCIITab(self, tabs: int) -> None:
        """Decrement the ASCII tab level.

        Collective.

        See Also
        --------
        petsc.PetscViewerASCIISubtractTab

        """
        cdef PetscInt ctabs = asInt(tabs)
        CHKERR(PetscViewerASCIISubtractTab(self.vwr, ctabs))

    def pushASCIISynchronized(self) -> None:
        """Allow ASCII synchronized calls.

        Collective.

        See Also
        --------
        printfASCIISynchronized, popASCIISynchronized
        petsc.PetscViewerASCIIPushSynchronized

        """
        CHKERR(PetscViewerASCIIPushSynchronized(self.vwr))

    def popASCIISynchronized(self) -> None:
        """Disallow ASCII synchronized calls.

        Collective.

        See Also
        --------
        printfASCIISynchronized, pushASCIISynchronized
        petsc.PetscViewerASCIIPopSynchronized

        """
        CHKERR(PetscViewerASCIIPopSynchronized(self.vwr))

    def pushASCIITab(self) -> None:
        """Push an additional tab level.

        Collective.

        See Also
        --------
        popASCIITab, petsc.PetscViewerASCIIPushTab

        """
        CHKERR(PetscViewerASCIIPushTab(self.vwr))

    def popASCIITab(self) -> None:
        """Pop an additional tab level pushed via `pushASCIITab`.

        Collective.

        See Also
        --------
        pushASCIITab, petsc.PetscViewerASCIIPopTab

        """
        CHKERR(PetscViewerASCIIPopTab(self.vwr))

    def useASCIITabs(self, flag: bool) -> None:
        """Enable/disable the use of ASCII tabs.

        Collective.

        See Also
        --------
        petsc.PetscViewerASCIIUseTabs

        """
        cdef PetscBool flg = asBool(flag)
        CHKERR(PetscViewerASCIIUseTabs(self.vwr, flg))

    def printfASCII(self, msg: str) -> None:
        """Print a message.

        Collective.

        See Also
        --------
        petsc.PetscViewerASCIIPrintf

        """
        cdef const char *cmsg = NULL
        msg = str2bytes(msg, &cmsg)
        CHKERR(PetscViewerASCIIPrintf(self.vwr, '%s', cmsg))

    def printfASCIISynchronized(self, msg: str) -> None:
        """Print a synchronized message.

        Collective.

        See Also
        --------
        pushASCIISynchronized, petsc.PetscViewerASCIISynchronizedPrintf

        """
        cdef const char *cmsg = NULL
        msg = str2bytes(msg, &cmsg)
        CHKERR(PetscViewerASCIISynchronizedPrintf(self.vwr, '%s', cmsg))

    # --- methods specific to file viewers ---

    def flush(self) -> None:
        """Flush the viewer.

        Collective.

        See Also
        --------
        petsc.PetscViewerFlush

        """
        CHKERR(PetscViewerFlush(self.vwr))

    def setFileMode(self, mode: FileMode | str) -> None:
        """Set file mode.

        Collective.

        See Also
        --------
        getFileMode, petsc.PetscViewerFileSetMode

        """
        CHKERR(PetscViewerFileSetMode(self.vwr, filemode(mode)))

    def getFileMode(self) -> FileMode:
        """Return the file mode.

        Not collective.

        See Also
        --------
        setFileMode, petsc.PetscViewerFileGetMode

        """
        cdef PetscFileMode mode = PETSC_FILE_MODE_READ
        CHKERR(PetscViewerFileGetMode(self.vwr, &mode))
        return mode

    def setFileName(self, name: str) -> None:
        """Set file name.

        Collective.

        See Also
        --------
        getFileName, petsc.PetscViewerFileSetName

        """
        cdef const char *cval = NULL
        name = str2bytes(name, &cval)
        CHKERR(PetscViewerFileSetName(self.vwr, cval))

    def getFileName(self) -> str:
        """Return file name.

        Not collective.

        See Also
        --------
        setFileName, petsc.PetscViewerFileGetName

        """
        cdef const char *cval = NULL
        CHKERR(PetscViewerFileGetName(self.vwr, &cval))
        return bytes2str(cval)

    # --- methods specific to draw viewers ---

    def setDrawInfo(
        self,
        display: str | None = None,
        title: str | None = None,
        position: tuple[int, int] | None = None,
        size: tuple[int, int] | int | None = None,
       ) -> None:
        """Set window information for a `Type.DRAW` viewer.

        Collective.

        Parameters
        ----------
        display
            The X display to use or `None` for the local machine.
        title
            The window title or `None` for no title.
        position
            Screen coordinates of the upper left corner, or `None` for default.
        size
            Window size or `None` for default.

        """
        # FIXME missing manual page
        # See Also
        # --------
        # petsc.PetscViewerDrawSetInfo
        cdef const char *cdisplay = NULL
        cdef const char *ctitle = NULL
        display = str2bytes(display, &cdisplay)
        title = str2bytes(title, &ctitle)
        cdef int x, y, h, w
        x = y = h = w = PETSC_DECIDE
        if position not in (None, PETSC_DECIDE):
            x, y = position
        if size not in (None, PETSC_DECIDE):
            try:
                w, h = size
            except TypeError:
                w = h = size
        CHKERR(PetscViewerDrawSetInfo(self.vwr,
                                      cdisplay, ctitle,
                                      x, y, w, h))

    def clearDraw(self) -> None:
        """Reset graphics.

        Not collective.

        See Also
        --------
        petsc.PetscViewerDrawClear

        """
        CHKERR(PetscViewerDrawClear(self.vwr))

    def createPython(self, context: Any = None, comm: Comm | None = None) -> Self:
        """Create a `Type.PYTHON` viewer.

        Collective.

        Parameters
        ----------
        context
            An instance of the Python class implementing the required methods.
        comm
            MPI communicator, defaults to `Sys.getDefaultComm`.

        See Also
        --------
        petsc_python_viewer, setType, setPythonContext, Type.PYTHON

        """
        # communicator and sizes
        cdef MPI_Comm ccomm = def_Comm(comm, PETSC_COMM_DEFAULT)
        cdef PetscViewer newvwr = NULL
        CHKERR(PetscViewerCreate(ccomm, &newvwr))
        CHKERR(PetscCLEAR(self.obj)); self.vwr = newvwr
        CHKERR(PetscViewerSetType(self.vwr, PETSCVIEWERPYTHON))
        CHKERR(PetscViewerPythonSetContext(self.vwr, <void*>context))
        if context:
            CHKERR(PetscViewerSetUp(self.vwr))
        return self

    def setPythonContext(self, context: Any) -> None:
        """Set the instance of the class implementing the required Python methods.

        Logically collective.

        See Also
        --------
        petsc_python_viewer, getPythonContext, setPythonType

        """
        CHKERR(PetscViewerPythonSetContext(self.vwr, <void*>context))

    def getPythonContext(self) -> Any:
        """Return the instance of the class implementing the required Python methods.

        Not collective.

        See Also
        --------
        petsc_python_viewer, setPythonContext

        """
        cdef void *context = NULL
        CHKERR(PetscViewerPythonGetContext(self.vwr, &context))
        if context == NULL: return None
        else: return <object> context

    def setPythonType(self, py_type: str) -> None:
        """Set the fully qualified Python name of the class to be used.

        Collective.

        See Also
        --------
        petsc_python_viewer, setPythonContext, getPythonType
        petsc.PetscViewerPythonSetType

        """
        cdef const char *cval = NULL
        py_type = str2bytes(py_type, &cval)
        CHKERR(PetscViewerPythonSetType(self.vwr, cval))

    def getPythonType(self) -> str:
        """Return the fully qualified Python name of the class used by the viewer.

        Not collective.

        See Also
        --------
        petsc_python_viewer, setPythonContext, setPythonType
        petsc.PetscViewerPythonGetType

        """
        cdef const char *cval = NULL
        CHKERR(PetscViewerPythonGetType(self.vwr, &cval))
        return bytes2str(cval)

    def viewObjectPython(self, Object obj) -> None:
        """View a generic `Object`.

        Collective.

        See Also
        --------
        petsc_python_viewer, setPythonContext, setPythonType
        petsc.PetscViewerPythonViewObject

        """
        CHKERR(PetscViewerPythonViewObject(self.vwr, obj.obj[0]))

# --------------------------------------------------------------------

cdef class ViewerHDF5(Viewer):
    """Viewer object for HDF5 file formats.

    Viewer is described in the `PETSc manual <petsc:sec_viewers>`.

    See Also
    --------
    Viewer

    """

    def create(
        self,
        name: str,
        mode: Viewer.FileMode | str | None = None,
        comm: Comm | None = None,
       ) -> Self:
        """Create a viewer of type `Type.HDF5`.

        Collective.

        Parameters
        ----------
        name
            The filename associated with the viewer.
        mode
            The mode type.
        comm
            MPI communicator, defaults to `Sys.getDefaultComm`.

        See Also
        --------
        Viewer.createHDF5

        """
        cdef MPI_Comm ccomm = def_Comm(comm, PETSC_COMM_DEFAULT)
        cdef const char *cname = NULL
        name = str2bytes(name, &cname)
        cdef PetscFileMode cmode = filemode(mode)
        cdef PetscViewer newvwr = NULL
        CHKERR(PetscViewerCreate(ccomm, &newvwr))
        CHKERR(PetscCLEAR(self.obj)); self.vwr = newvwr
        CHKERR(PetscViewerSetType(self.vwr, PETSCVIEWERHDF5))
        CHKERR(PetscViewerFileSetMode(self.vwr, cmode))
        CHKERR(PetscViewerFileSetName(self.vwr, cname))
        return self

    def pushTimestepping(self) -> None:
        """Activate the timestepping mode.

        Logically collective.

        See Also
        --------
        popTimestepping, petsc.PetscViewerHDF5PushTimestepping

        """
        CHKERR(PetscViewerHDF5PushTimestepping(self.vwr))

    def popTimestepping(self) -> None:
        """Deactivate the timestepping mode.

        Logically collective.

        See Also
        --------
        pushTimestepping, petsc.PetscViewerHDF5PopTimestepping

        """
        CHKERR(PetscViewerHDF5PopTimestepping(self.vwr))

    def getTimestep(self) -> int:
        """Return the current time step.

        Not collective.

        See Also
        --------
        pushTimestepping, setTimestep, incrementTimestep
        petsc.PetscViewerHDF5GetTimestep

        """
        cdef PetscInt ctimestep = 0
        CHKERR(PetscViewerHDF5GetTimestep(self.vwr, &ctimestep))
        return toInt(ctimestep)

    def setTimestep(self, timestep: int) -> None:
        """Set the current time step.

        Logically collective.

        See Also
        --------
        pushTimestepping, getTimestep, incrementTimestep
        petsc.PetscViewerHDF5SetTimestep

        """
        CHKERR(PetscViewerHDF5SetTimestep(self.vwr, asInt(timestep)))

    def incrementTimestep(self) -> None:
        """Increment the time step.

        Logically collective.

        See Also
        --------
        pushTimestepping, setTimestep, getTimestep
        petsc.PetscViewerHDF5IncrementTimestep

        """
        CHKERR(PetscViewerHDF5IncrementTimestep(self.vwr))

    def pushGroup(self, group: str) -> None:
        """Set the current group.

        Logically collective.

        See Also
        --------
        popGroup, getGroup, petsc.PetscViewerHDF5PushGroup

        """
        cdef const char *cgroup = NULL
        group = str2bytes(group, &cgroup)
        CHKERR(PetscViewerHDF5PushGroup(self.vwr, cgroup))

    def popGroup(self) -> None:
        """Pop the current group from the stack.

        Logically collective.

        See Also
        --------
        pushGroup, getGroup, petsc.PetscViewerHDF5PopGroup

        """
        CHKERR(PetscViewerHDF5PopGroup(self.vwr))

    def getGroup(self) -> str:
        """Return the current group.

        Not collective.

        See Also
        --------
        pushGroup, popGroup, petsc.PetscViewerHDF5GetGroup

        """
        cdef const char *cgroup = NULL
        CHKERR(PetscViewerHDF5GetGroup(self.vwr, NULL, &cgroup))
        group = bytes2str(cgroup)
        CHKERR(PetscFree(cgroup))
        return group

# --------------------------------------------------------------------

del ViewerType
del ViewerFormat
del ViewerFileMode
del ViewerDrawSize

# --------------------------------------------------------------------



# create linear solver
ksp = PETSc.KSP()
ksp.create(PETSc.COMM_WORLD)
# use conjugate gradients
ksp.setType('cg')
# and incomplete Cholesky
ksp.getPC().setType('icc')
# obtain sol & rhs vectors
x, b = A.createVecs()
x.set(0)
b.set(1)
# and next solve
ksp.setOperators(A)
ksp.setFromOptions()
ksp.solve(b, x)



# DLPack interface

cdef extern from "Python.h":
    ctypedef void (*PyCapsule_Destructor)(object)
    bint PyCapsule_IsValid(object, const char*)
    void* PyCapsule_GetPointer(object, const char*) except? NULL
    int PyCapsule_SetName(object, const char*) except -1
    object PyCapsule_New(void*, const char*, PyCapsule_Destructor)
    int PyCapsule_CheckExact(object)

cdef extern from "<stdlib.h>" nogil:
    ctypedef signed long int64_t
    ctypedef unsigned long long uint64_t
    ctypedef unsigned char uint8_t
    ctypedef unsigned short uint16_t
    void free(void* ptr)
    void* malloc(size_t size)

cdef struct DLDataType:
    uint8_t code
    uint8_t bits
    uint16_t lanes

cdef enum PetscDLDeviceType:
    kDLCPU = <unsigned int>1
    kDLCUDA = <unsigned int>2
    kDLCUDAHost = <unsigned int>3
    # kDLOpenCL = <unsigned int>4
    # kDLVulkan = <unsigned int>7
    # kDLMetal = <unsigned int>8
    # kDLVPI = <unsigned int>9
    kDLROCM = <unsigned int>10
    kDLROCMHost = <unsigned int>11
    # kDLExtDev = <unsigned int>12
    kDLCUDAManaged = <unsigned int>13
    # kDLOneAPI = <unsigned int>14

ctypedef struct DLContext:
    PetscDLDeviceType device_type
    int device_id

cdef enum DLDataTypeCode:
    kDLInt = <unsigned int>0
    kDLUInt = <unsigned int>1
    kDLFloat = <unsigned int>2

cdef struct DLTensor:
    void* data
    DLContext ctx
    int ndim
    DLDataType dtype
    int64_t* shape
    int64_t* strides
    uint64_t byte_offset

ctypedef int (*dlpack_manager_del_obj)(void*) noexcept nogil

cdef struct DLManagedTensor:
    DLTensor dl_tensor
    void* manager_ctx
    void (*manager_deleter)(DLManagedTensor*) noexcept nogil
    dlpack_manager_del_obj del_obj

cdef void pycapsule_deleter(object dltensor) noexcept:
    cdef DLManagedTensor* dlm_tensor = NULL
    # we do not call a used capsule's deleter
    if PyCapsule_IsValid(dltensor, b'dltensor'):
        dlm_tensor = <DLManagedTensor *>PyCapsule_GetPointer(dltensor, b'dltensor')
        manager_deleter(dlm_tensor)

cdef void manager_deleter(DLManagedTensor* tensor) noexcept nogil:
    if tensor.manager_ctx is NULL:
        return
    free(tensor.dl_tensor.shape)
    if tensor.del_obj is not NULL:
        tensor.del_obj(&tensor.manager_ctx)
    free(tensor)

# --------------------------------------------------------------------
